# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""Utility methods used only by the CLI client."""

import click
from pyvcloud.vcd.client import Client
import pyvcloud.vcd.org as vcd_org
import pyvcloud.vcd.utils as vcd_utils
import requests
import semantic_version
import six
from vcd_cli.profiles import Profiles
from vcd_cli.utils import stdout

from container_service_extension.client import system as syst
from container_service_extension.client.constants import CSE_SERVER_RUNNING
import container_service_extension.common.constants.shared_constants as shared_constants  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_SERVER_API_VERSION  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_SERVER_BUSY_KEY  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_SERVER_LEGACY_MODE  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_SERVER_SUPPORTED_API_VERSIONS  # noqa: E501
from container_service_extension.common.utils.core_utils import extract_id_from_href  # noqa: E501
from container_service_extension.exception.exceptions import CseResponseError
from container_service_extension.logging.logger import CLIENT_LOGGER
from container_service_extension.logging.logger import NULL_LOGGER
import container_service_extension.rde.constants as def_constants

_RESTRICT_CLI_TO_TKG_S_OPERATIONS = False


def is_cli_for_tkg_s_only():
    global _RESTRICT_CLI_TO_TKG_S_OPERATIONS
    return _RESTRICT_CLI_TO_TKG_S_OPERATIONS


def restrict_cli_to_tkg_s_operations():
    global _RESTRICT_CLI_TO_TKG_S_OPERATIONS
    _RESTRICT_CLI_TO_TKG_S_OPERATIONS = True


def enable_cli_for_all_operations():
    global _RESTRICT_CLI_TO_TKG_S_OPERATIONS
    _RESTRICT_CLI_TO_TKG_S_OPERATIONS = False


def cse_restore_session(ctx) -> None:
    """Restores the session with vcd client with right server api version.

    Replace the vcd client in ctx.obj with new client created with server api
    version. Also saves the server api version in profiles.

    :param <click.core.Context> ctx: click context

    :return:
    """
    # Always override the vcd_client by new client with CSE server api version.
    if type(ctx.obj) is not dict or not ctx.obj.get('client'):
        CLIENT_LOGGER.debug('Restoring client from profile.')
        profiles = Profiles.load()
        token = profiles.get('token')
        if token is None or len(token) == 0:
            msg = "Can't restore session, please login again."
            CLIENT_LOGGER.debug(f"Missing Token : {msg}")
            raise Exception(msg)
        if not profiles.get('verify'):
            if not profiles.get('disable_warnings'):
                click.secho(
                    'InsecureRequestWarning: '
                    'Unverified HTTPS request is being made. '
                    'Adding certificate verification is strongly '
                    'advised.',
                    fg='yellow',
                    err=True)
            requests.packages.urllib3.disable_warnings()

        client = Client(
            profiles.get('host'),
            api_version=profiles.get('api_version'),
            verify_ssl_certs=profiles.get('verify'),
            log_file='vcd.log',
            log_requests=profiles.get('log_request'),
            log_headers=profiles.get('log_header'),
            log_bodies=profiles.get('log_body'))
        client.rehydrate_from_token(
            profiles.get('token'), profiles.get('is_jwt_token'))

        ctx.obj = {'client': client}
    else:
        CLIENT_LOGGER.debug('Reusing client from context.')

    _override_client(ctx)


def _override_client(ctx) -> None:
    """Replace the vcd client in ctx.obj with new one.

    New vcd client takes the CSE server_api_version as api_version param.
    Save profile also with 'cse_server_api_version' for subsequent commands.

    :param <click.core.Context> ctx: click context
    """
    profiles = Profiles.load()
    # if the key CSE_SERVER_RUNNING is not present in the profiles.yaml,
    # we make an assumption that CSE server is running
    is_cse_server_running = profiles.get(CSE_SERVER_RUNNING, default=True)
    cse_server_api_version = profiles.get(CSE_SERVER_API_VERSION)
    if not is_cse_server_running:
        CLIENT_LOGGER.debug("CSE server not running as per profile, restricting CLI to only TKG operations.")  # noqa: E501
        restrict_cli_to_tkg_s_operations()
        ctx.obj['profiles'] = profiles
        return

    # Get server_api_version; save it in profiles if doesn't exist
    if not cse_server_api_version:
        try:
            system = syst.System(ctx.obj['client'])
            sys_info = system.get_info()

            is_pre_cse_3_1_server = False
            if CSE_SERVER_LEGACY_MODE not in sys_info or \
                    CSE_SERVER_SUPPORTED_API_VERSIONS not in sys_info:
                is_pre_cse_3_1_server = True

            if not is_pre_cse_3_1_server:
                is_cse_server_running_in_legacy_mode = \
                    sys_info.get(CSE_SERVER_LEGACY_MODE)
                cse_server_supported_api_versions = \
                    set(sys_info.get(CSE_SERVER_SUPPORTED_API_VERSIONS))
                cse_client_supported_api_versions = \
                    set(shared_constants.SUPPORTED_VCD_API_VERSIONS)

                common_supported_api_versions = \
                    list(cse_server_supported_api_versions.intersection(
                        cse_client_supported_api_versions))

                # ToDo: Instead of float use proper version comparison
                if is_cse_server_running_in_legacy_mode:
                    common_supported_api_versions = \
                        [float(x) for x in common_supported_api_versions
                            if float(x) < 35.0]
                else:
                    common_supported_api_versions = \
                        [float(x) for x in common_supported_api_versions
                            if float(x) >= 35.0]

                cse_server_api_version = \
                    str(max(common_supported_api_versions))
                CLIENT_LOGGER.debug(
                    f"Server api versions : {cse_server_supported_api_versions}, "  # noqa: E501
                    f"Client api versions : {cse_client_supported_api_versions}, "  # noqa: E501
                    f"Server in Legacy mode : {is_cse_server_running_in_legacy_mode}, "  # noqa: E501
                    f"Selected api version : {cse_server_api_version}."
                )
            else:
                cse_server_api_version = \
                    sys_info.get(CSE_SERVER_API_VERSION)
                CLIENT_LOGGER.debug(
                    "Pre CSE 3.1 server detected. Selected api version : "
                    f"{cse_server_api_version}.")

            profiles.set(CSE_SERVER_API_VERSION, cse_server_api_version)
            profiles.set(CSE_SERVER_RUNNING, True)
            profiles.save()
        except (requests.exceptions.Timeout, CseResponseError) as err:
            CLIENT_LOGGER.error(err, exc_info=True)
            CLIENT_LOGGER.debug("Request to CSE server timed out. Restricting CLI to only TKG operations.")  # noqa: E501

            profiles.set(CSE_SERVER_RUNNING, False)
            restrict_cli_to_tkg_s_operations()
            ctx.obj['profiles'] = profiles
            profiles.save()
            return
    client = Client(
        profiles.get('host'),
        api_version=cse_server_api_version,
        verify_ssl_certs=profiles.get('verify'),
        log_file='vcd.log',
        log_requests=profiles.get('log_request'),
        log_headers=profiles.get('log_header'),
        log_bodies=profiles.get('log_body'))
    client.rehydrate_from_token(profiles.get('token'), profiles.get('is_jwt_token'))  # noqa: E501
    ctx.obj['client'] = client
    ctx.obj['profiles'] = profiles


def construct_filters(server_rde_version, **kwargs):
    # NOTE: org and vdc filters need to be camel cased if RDE version > 1.0.0
    filter_key = def_constants.ClusterEntityFilterKey2X
    if semantic_version.Version(server_rde_version) <= \
            semantic_version.Version(def_constants.RDEVersion.RDE_1_0_0):
        filter_key = def_constants.ClusterEntityFilterKey1X
    filters = {}
    if kwargs.get('org'):
        filters[filter_key.ORG_NAME.value] = kwargs['org']  # noqa: E501
    if kwargs.get('vdc'):
        filters[filter_key.OVDC_NAME.value] = kwargs['vdc']  # noqa: E501
    return filters


def construct_task_console_message(task_href: str) -> str:
    msg = "Run the following command to track the status of the cluster:\n"
    task_id = task_href.split('/')[-1]
    msg += f"vcd task wait {task_id}"
    return msg


def swagger_object_to_dict(obj):
    """Convert a swagger object to a dictionary without changing case type."""
    # reference: https://github.com/swagger-api/swagger-codegen/issues/8948
    result = {}
    o_map = obj.attribute_map

    for attr, _ in six.iteritems(obj.swagger_types):
        value = getattr(obj, attr)
        if isinstance(value, list):
            result[o_map[attr]] = list(map(
                lambda x: swagger_object_to_dict(x) if hasattr(x, "to_dict") else x,  # noqa: E501
                value
            ))
        elif hasattr(value, "to_dict"):
            result[o_map[attr]] = swagger_object_to_dict(value)
        elif isinstance(value, dict):
            result[o_map[attr]] = dict(map(
                lambda item: (item[0], swagger_object_to_dict(item[1]))
                if hasattr(item[1], "to_dict") else item,
                value.items()
            ))
        else:
            result[o_map[attr]] = value

    return result


def filter_columns(result, value_field_to_display_field):
    """Extract selected fields from each list item in result.

    :param list(dict) or dict result: row of records
    :param dict value_field_to_display_field: mapping of value field to
    respective display name in result set. Extract selected fields from result
    based on value fields from this dictionary.
    :return: filtered list or dict of record(s) from result
    :rtype: list(dict) or dict
    """
    if isinstance(result, list):
        filtered_result = []
        for result_record in result:
            filtered_record = {}
            for value_field, display_field in value_field_to_display_field.items():  # noqa: E501
                filtered_record[display_field] = result_record.get(value_field, '')  # noqa: E501
            filtered_result.append(filtered_record)
        return filtered_result
    elif isinstance(result, dict):
        # If the server result is the CSE Server busy message, there is no
        # result to filter, so the CSE Server busy message is returned as is
        if result.get(CSE_SERVER_BUSY_KEY) is not None:
            return result

        filtered_result = {
            display_field: result.get(value_field, '')
            for value_field, display_field in value_field_to_display_field.items()  # noqa: E501
        }
        return filtered_result


def create_user_name_to_id_dict(client: Client, users_set: set, org_href):
    """Get a dictionary of users to user ids from a list of user names.

    :param Client client: current client
    :param set users_set: set of user names
    :param str org_href: href of the org to search in

    :return: dict of user name keys and user id values
    :rtype: dict
    :raise Exception is not all id's are found for users
    """
    own_users_set = users_set.copy()
    org = vcd_org.Org(client, org_href)
    org_users = org.list_users()
    user_name_to_id_dict = {}
    for user_str_elem in org_users:
        curr_user_dict = vcd_utils.to_dict(user_str_elem, exclude=[])
        curr_user_name = curr_user_dict['name']
        if curr_user_name in own_users_set:
            user_id = extract_id_from_href(curr_user_dict['href'])
            user_name_to_id_dict[curr_user_name] = shared_constants.USER_URN_PREFIX + user_id  # noqa: E501
            own_users_set.remove(curr_user_name)

        # Stop searching if all needed names and ids found
        if len(own_users_set) == 0:
            break
    if len(own_users_set) > 0:
        raise Exception(f"No user ids found for: {list(own_users_set)}")
    return user_name_to_id_dict


def access_level_reduced(new_access_urn, curr_access_urn):
    """Check if the access level is reduced.

    Only true is the current access level is full-control and the new access
    level is not full control, or is the current access level is read-write and
    the new access level is read-only.
    """
    if curr_access_urn == shared_constants.FULL_CONTROL_ACCESS_LEVEL_ID and \
            new_access_urn != shared_constants.FULL_CONTROL_ACCESS_LEVEL_ID:
        return True
    if curr_access_urn == shared_constants.READ_WRITE_ACCESS_LEVEL_ID and \
            new_access_urn == shared_constants.READ_ONLY_ACCESS_LEVEL_ID:
        return True
    return False


def print_paginated_result(generator, should_print_all=False, logger=NULL_LOGGER):  # noqa: E501
    """Print results by prompting the user for more results.

    :param Generator[(List[dict], int), None, None] generator: generator which
        yields a list of results and a boolean indicating if more results
        are present.
    :param bool should_print_all: print all the results without prompting the
        user.
    :param logger logger: logger to log the results or exceptions.
    """
    try:
        headers_printed = False
        for result, has_more_results in generator:
            stdout(result, sort_headers=False,
                   show_headers=not headers_printed)
            headers_printed = True
            logger.debug(result)
            if not has_more_results or \
                    not (should_print_all or click.confirm("Do you want more results?")):  # noqa: E501
                break
    except Exception as e:
        logger.error(f"Error while iterating over the paginated response: {e}")
        raise



# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""Support CSE 3.1 cluster requests.

Starting CSE 3.1, this module handles all the cluster requests coming in at
any RDE version (>= 2.0) and at any request api_version (>=36).

Responsibility of the functions in this file:
1. Validate the Input payload based on the (Operation, payload_version).
   Get the validator based on the payload_version.
2. Convert the input entity to runtime rde format.
3. Based on the runtime rde, call the appropriate backend method.
4. Convert the response to be compatible with the request payload
   version (If request = X.Y, the response must be X.Z where Z>=Y)
5. Return the response.
"""

import container_service_extension.common.constants.server_constants as server_constants  # noqa: E501
from container_service_extension.common.constants.shared_constants import ClusterAclKey  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_PAGINATION_DEFAULT_PAGE_SIZE  # noqa: E501
from container_service_extension.common.constants.shared_constants import CSE_PAGINATION_FIRST_PAGE_NUMBER  # noqa: E501
from container_service_extension.common.constants.shared_constants import PaginationKey  # noqa: E501
from container_service_extension.common.constants.shared_constants import RequestKey  # noqa: E501
import container_service_extension.common.thread_local_data as thread_local_data  # noqa: E501
import container_service_extension.common.utils.pyvcloud_utils as pyvcloud_utils  # noqa: E501
import container_service_extension.common.utils.server_utils as server_utils
import container_service_extension.lib.telemetry.constants as telemetry_constants  # noqa: E501
import container_service_extension.lib.telemetry.telemetry_handler as telemetry_handler  # noqa: E501
import container_service_extension.rde.backend.cluster_service_factory as cluster_service_factory  # noqa: E501
from container_service_extension.rde.behaviors.behavior_model import BehaviorOperation  # noqa: E501
from container_service_extension.rde.behaviors.behavior_model import DELETE_NFS_NODE_BEHAVIOR_ID  # noqa: E501
from container_service_extension.rde.behaviors.behavior_model import KUBE_CONFIG_BEHAVIOR_INTERFACE_ID  # noqa: E501
from container_service_extension.rde.behaviors.behavior_service import BehaviorService  # noqa: E501
import container_service_extension.rde.common.entity_service as entity_service
import container_service_extension.rde.constants as rde_constants
from container_service_extension.rde.models.abstractNativeEntity import AbstractNativeEntity  # noqa: E501
import container_service_extension.rde.models.common_models as common_models
import container_service_extension.rde.utils as rde_utils
import container_service_extension.rde.validators.validator_factory as rde_validator_factory  # noqa: E501
import container_service_extension.security.context.behavior_request_context as behavior_ctx  # noqa: E501
import container_service_extension.security.context.operation_context as ctx
import container_service_extension.server.request_handlers.request_utils as request_utils  # noqa: E501

# TODO: Needs to evaluate if we want to handle RDE 1.0 requests coming
#  in at 36.0 in this handler.


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V36_CLUSTER_APPLY)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_create(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster create operation.

    :return: Defined entity of the native cluster
    :rtype: container_service_extension.def_.models.DefEntity
    """
    input_entity: dict = data[RequestKey.INPUT_SPEC]
    payload_version = input_entity.get(rde_constants.PayloadKey.PAYLOAD_VERSION)  # noqa: E501
    rde_utils.raise_error_if_unsupported_payload_version(payload_version)

    # Validate the Input payload based on the (Operation, payload_version).
    # Get the validator based on the payload_version
    # ToDo : Don't use default cloudapi_client. Use the specific versioned one
    rde_validator_factory.get_validator(
        rde_version=rde_constants.MAP_INPUT_PAYLOAD_VERSION_TO_RDE_VERSION[
            payload_version]).validate(cloudapi_client=op_ctx.cloudapi_client,
                                       entity=input_entity)

    def_entity_service = entity_service.DefEntityService(op_ctx.cloudapi_client)  # noqa: E501
    entity_type = server_utils.get_registered_def_entity_type()
    converted_entity: AbstractNativeEntity = rde_utils.convert_input_rde_to_runtime_rde_format(input_entity)  # noqa: E501
    def_entity = common_models.DefEntity(entity=converted_entity, entityType=entity_type.id)  # noqa: E501

    # No need to set org context for non sysadmin users
    org_context = None
    if op_ctx.client.is_sysadmin():
        org_resource = pyvcloud_utils.get_org(op_ctx.client, org_name=converted_entity.metadata.org_name)  # noqa: E501
        org_context = org_resource.href.split('/')[-1]

    _, task_href = def_entity_service.create_entity(
        entity_type_id=entity_type.id,
        entity=def_entity,
        tenant_org_context=org_context,
        is_request_async=True)

    task_resource = op_ctx.sysadmin_client.get_resource(task_href)
    entity_id = task_resource.Owner.get('id')
    def_entity = def_entity_service.get_entity(entity_id)
    def_entity.entity.status.task_href = task_href
    return def_entity.to_dict()


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V36_CLUSTER_LIST)  # noqa: E501
@request_utils.cluster_api_exception_handler
def native_cluster_list(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster list operation.

    :return: List
    """
    svc = cluster_service_factory.ClusterServiceFactory(_get_request_context(op_ctx)).get_cluster_service()  # noqa: E501
    filters = data.get(RequestKey.QUERY_PARAMS, {})
    page_number = int(filters.get(PaginationKey.PAGE_NUMBER,
                                  CSE_PAGINATION_FIRST_PAGE_NUMBER))
    page_size = int(filters.get(PaginationKey.PAGE_SIZE,
                                CSE_PAGINATION_DEFAULT_PAGE_SIZE))

    # remove page number and page size from the filters as it is treated
    # differently from other filters
    if PaginationKey.PAGE_NUMBER in filters:
        del filters[PaginationKey.PAGE_NUMBER]
    if PaginationKey.PAGE_SIZE in filters:
        del filters[PaginationKey.PAGE_SIZE]

    # response needs to paginated
    result = svc.get_clusters_by_page(filters=filters)
    clusters = [def_entity.to_dict() for def_entity in result[PaginationKey.VALUES]]  # noqa: E501

    uri = data['url']
    return server_utils.create_links_and_construct_paginated_result(
        uri,
        clusters,
        result_total=result[PaginationKey.RESULT_TOTAL],
        page_number=page_number,
        page_size=page_size,
        query_params=filters)


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V36_CLUSTER_LIST)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_list(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster list operation.

    :return: List
    """
    svc = cluster_service_factory.ClusterServiceFactory(_get_request_context(op_ctx)).get_cluster_service()  # noqa: E501
    # response should not be paginated
    return [def_entity.to_dict() for def_entity in
            svc.list_clusters(data.get(RequestKey.QUERY_PARAMS, {}))]


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V36_CLUSTER_INFO)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_info(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster info operation.

    Required data: cluster_name
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    svc = cluster_service_factory.ClusterServiceFactory(_get_request_context(op_ctx)).get_cluster_service()  # noqa: E501
    cluster_id = data[RequestKey.CLUSTER_ID]
    return svc.get_cluster_info(cluster_id).to_dict()


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V36_CLUSTER_CONFIG)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_config(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster config operation.

    Required data: cluster_id

    :return: Dict
    """
    cluster_id = data[RequestKey.CLUSTER_ID]
    behavior_svc = BehaviorService(cloudapi_client=op_ctx.cloudapi_client)
    config_task_href = behavior_svc.invoke_behavior(
        entity_id=cluster_id,
        behavior_interface_id=KUBE_CONFIG_BEHAVIOR_INTERFACE_ID)
    config_task = op_ctx.client.get_resource(config_task_href)
    op_ctx.client.get_task_monitor().wait_for_success(config_task)
    config_task = op_ctx.client.get_resource(config_task_href)
    return config_task.Result.ResultContent.text


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V36_CLUSTER_APPLY)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_update(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster resize operation.

    :return: Defined entity of the native cluster
    :rtype: container_service_extension.def_.models.DefEntity
    """
    cluster_id = data[RequestKey.CLUSTER_ID]
    input_entity: dict = data[RequestKey.INPUT_SPEC]
    payload_version = input_entity.get(rde_constants.PayloadKey.PAYLOAD_VERSION)  # noqa: E501
    rde_utils.raise_error_if_unsupported_payload_version(payload_version)

    # Validate the Input payload based on the (Operation, payload_version).
    # Get the validator based on the payload_version
    # ToDo : Don't use default cloudapi_client. Use the specific versioned one
    rde_validator_factory.get_validator(
        rde_version=rde_constants.MAP_INPUT_PAYLOAD_VERSION_TO_RDE_VERSION[
            payload_version]). \
        validate(cloudapi_client=op_ctx.cloudapi_client, entity_id=cluster_id,
                 entity=input_entity,
                 operation=BehaviorOperation.UPDATE_CLUSTER)

    # Convert the input entity to runtime rde format.
    # Based on the runtime rde, call the appropriate backend method.
    def_entity_service = entity_service.DefEntityService(op_ctx.cloudapi_client)  # noqa: E501
    converted_native_entity: AbstractNativeEntity = rde_utils.convert_input_rde_to_runtime_rde_format(input_entity)  # noqa: E501
    cluster_def_entity: common_models.DefEntity = def_entity_service.get_entity(cluster_id)  # noqa: E501
    cluster_def_entity.entity.spec = converted_native_entity.spec
    updated_def_entity, task_href = def_entity_service.update_entity(
        entity_id=cluster_id, entity=cluster_def_entity,
        invoke_hooks=True, is_request_async=True)
    updated_def_entity.entity.status.task_href = task_href
    # TODO: Response RDE must be compatible with the request RDE.
    #  Conversions may be needed especially if there is a major version
    #  difference in the input RDE and runtime RDE.
    return updated_def_entity.to_dict()


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V36_CLUSTER_UPGRADE_PLAN)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_upgrade_plan(data, op_ctx: ctx.OperationContext):
    """Request handler for cluster upgrade-plan operation.

    :return: List[Tuple(str, str)]
    """
    svc = cluster_service_factory.ClusterServiceFactory(_get_request_context(op_ctx)).get_cluster_service()  # noqa: E501
    return svc.get_cluster_upgrade_plan(data[RequestKey.CLUSTER_ID])


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V36_CLUSTER_DELETE)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_delete(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster delete operation.

    Required data: cluster_name
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    cluster_id = data[RequestKey.CLUSTER_ID]
    def_entity_service = entity_service.DefEntityService(op_ctx.cloudapi_client)  # noqa: E501
    def_entity: common_models.DefEntity = def_entity_service.get_entity(cluster_id)  # noqa: E501
    _, task_href = def_entity_service.delete_entity(cluster_id, invoke_hooks=True, is_request_async=True)  # noqa: E501
    def_entity.entity.status.task_href = task_href
    return def_entity.to_dict()


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V36_NODE_DELETE)  # noqa: E501
@request_utils.cluster_api_exception_handler
def nfs_node_delete(data, op_ctx: ctx.OperationContext):
    """Request handler for node delete operation.

    Required data: cluster_name, node_names_list
    Optional data and default values: org_name=None, ovdc_name=None

    (data validation handled in broker)

    :return: Dict
    """
    cluster_id = data[RequestKey.CLUSTER_ID]
    node_name = data[RequestKey.NODE_NAME]
    behavior_svc = BehaviorService(cloudapi_client=op_ctx.cloudapi_client)
    telemetry_handler.record_user_action_details(
        cse_operation=telemetry_constants.CseOperation.V36_NODE_DELETE,
        cse_params={
            telemetry_constants.PayloadKey.CLUSTER_ID: cluster_id,
            telemetry_constants.PayloadKey.NODE_NAME: node_name,
            telemetry_constants.PayloadKey.SOURCE_DESCRIPTION: thread_local_data.get_thread_local_data(server_constants.ThreadLocalData.USER_AGENT)   # noqa: E501
        }
    )
    delete_nfs_node_task = behavior_svc.invoke_behavior(
        entity_id=cluster_id,
        behavior_interface_id=DELETE_NFS_NODE_BEHAVIOR_ID,
        arguments={
            RequestKey.NODE_NAME.value: node_name
        })
    def_entity_service = entity_service.DefEntityService(op_ctx.cloudapi_client)  # noqa: E501
    cluster_def_entity: common_models.DefEntity = def_entity_service.get_entity(cluster_id)  # noqa: E501
    cluster_def_entity.entity.status.task_href = delete_nfs_node_task

    return cluster_def_entity.to_dict()


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V36_CLUSTER_ACL_LIST)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_acl_info(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster acl list operation."""
    svc = cluster_service_factory.ClusterServiceFactory(_get_request_context(op_ctx)).get_cluster_service()  # noqa: E501
    cluster_id = data[RequestKey.CLUSTER_ID]
    query = data.get(RequestKey.QUERY_PARAMS, {})
    page = int(query.get(PaginationKey.PAGE_NUMBER, CSE_PAGINATION_FIRST_PAGE_NUMBER))  # noqa: E501
    page_size = int(query.get(PaginationKey.PAGE_SIZE, CSE_PAGINATION_DEFAULT_PAGE_SIZE))  # noqa: E501
    result: dict = svc.get_cluster_acl_info(cluster_id, page, page_size)

    uri = data['url']
    return server_utils.create_links_and_construct_paginated_result(
        base_uri=uri,
        values=result.get(PaginationKey.VALUES, []),
        result_total=result.get(PaginationKey.RESULT_TOTAL, 0),
        page_number=page,
        page_size=page_size)


@telemetry_handler.record_user_action_telemetry(cse_operation=telemetry_constants.CseOperation.V36_CLUSTER_ACL_UPDATE)  # noqa: E501
@request_utils.cluster_api_exception_handler
def cluster_acl_update(data: dict, op_ctx: ctx.OperationContext):
    """Request handler for cluster acl update operation."""
    svc = cluster_service_factory.ClusterServiceFactory(_get_request_context(op_ctx)).get_cluster_service()  # noqa: E501
    cluster_id = data[RequestKey.CLUSTER_ID]
    update_acl_entries = data.get(RequestKey.INPUT_SPEC, {}).get(ClusterAclKey.ACCESS_SETTING)  # noqa: E501
    svc.update_cluster_acl(cluster_id, update_acl_entries)


def _get_request_context(op_ctx: ctx.OperationContext):
    """Get the request context from operation context.

    :param op_ctx: operation context
    :return: RequestContext
    :rtype: behavior_ctx.RequestContext
    """
    return behavior_ctx.RequestContext(op_ctx=op_ctx, mqtt_publisher=op_ctx.mqtt_publisher)  # noqa: E501



# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import semantic_version

from container_service_extension.rde.models.rde_1_0_0 import NativeEntity as NativeEntity1X  # noqa: E501
from container_service_extension.rde.models.rde_2_0_0 import NativeEntity as NativeEntity2X  # noqa: E501


def get_rde_model(rde_version):
    """Get the model class of the specified rde_version.

    Factory method to return the model class based on the specified RDE version
    :param rde_version (str)

    :rtype model: NativeEntity
    """
    rde_version: semantic_version.Version = semantic_version.Version(rde_version)  # noqa: E501
    if rde_version.major == 1:
        return NativeEntity1X
    elif rde_version.major == 2:
        return NativeEntity2X



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

from typing import Dict, Optional

import container_service_extension.common.utils.pyvcloud_utils as vcd_utils
import container_service_extension.common.utils.server_utils as server_utils
import container_service_extension.logging.logger as logger
import container_service_extension.security.context.user_context as user_context  # noqa: E501


class OperationContext:
    def __init__(self, auth_token: str, is_jwt: bool = True, request_id: Optional[str] = None, mqtt_publisher=None):  # noqa: E501
        self._auth_token: str = auth_token
        self._is_jwt: bool = is_jwt
        # Request ID; may be None if OperationContext is initialized outside of
        # request_dispatcher.py
        self.request_id: Optional[str] = request_id

        # map for storing user's context at different api versions
        # `None` key, maps to the client at the highest api version supported
        # by the vCD and pyvcloud
        self._user_context_map: Dict[Optional[str], user_context.UserContext] = {}  # noqa: E501

        # map for storing sys admin user's context at different api versions
        # `None` key maps to the client at the highest api version supported
        # by the vCD and pyvcloud
        self._sysadmin_user_context_map: Dict[Optional[str], user_context.UserContext] = {}  # noqa: E501

        # async operations should call end() when they are finished
        self.is_async: bool = False

        self.mqtt_publisher = mqtt_publisher

    @property
    def client(self):
        return self.user.client

    @property
    def cloudapi_client(self):
        return self.user.cloud_api_client

    @property
    def sysadmin_client(self):
        return self.sysadmin_user.client

    @property
    def sysadmin_cloudapi_client(self):
        return self.sysadmin_user.cloud_api_client

    @property
    def user(self):
        api_version = None  # marker for default api version
        return self.get_user_context(api_version)

    @property
    def sysadmin_user(self):
        api_version = None  # marker for default api version
        return self.get_sysadmin_user_context(api_version)

    def get_client(self, api_version: Optional[str]):
        return self.get_user_context(api_version).client

    def get_cloudapi_client(self, api_version: Optional[str]):
        return self.get_user_context(api_version).cloud_api_client

    def get_sysadmin_client(self, api_version: Optional[str]):
        return self.get_sysadmin_user_context(api_version).client

    def get_sysadmin_cloudapi_client(self, api_version: Optional[str]):
        return self.get_sysadmin_user_context(api_version).cloud_api_client

    def get_user_context(self, api_version: Optional[str]):
        if api_version not in self._user_context_map:
            self._update_user_context_map(api_version=api_version)
        return self._user_context_map[api_version]

    def get_sysadmin_user_context(self, api_version: Optional[str]):
        if api_version not in self._sysadmin_user_context_map:
            self._update_sysadmin_user_context_map(api_version=api_version)
        return self._sysadmin_user_context_map[api_version]

    def _update_user_context_map(self, api_version: Optional[str]):
        _client = vcd_utils.connect_vcd_user_via_token(
            tenant_auth_token=self._auth_token,
            is_jwt_token=self._is_jwt,
            api_version=api_version)

        log_wire = server_utils.get_server_runtime_config() \
            .get('service', {}).get('log_wire', False)  # noqa: E501
        logger_wire = logger.NULL_LOGGER
        if log_wire:
            logger_wire = logger.SERVER_CLOUDAPI_WIRE_LOGGER
        _cloudapi_client = vcd_utils.get_cloudapi_client_from_vcd_client(
            _client,
            logger.SERVER_LOGGER,
            logger_wire)

        _user_context = user_context.UserContext(
            client=_client, cloud_api_client=_cloudapi_client)
        self._user_context_map[api_version] = _user_context

    def _update_sysadmin_user_context_map(self, api_version: Optional[str]):
        _sysadmin_client = vcd_utils.get_sys_admin_client(
            api_version=api_version)

        log_wire = server_utils.get_server_runtime_config() \
            .get('service', {}).get('log_wire', False)
        logger_wire = logger.NULL_LOGGER
        if log_wire:
            logger_wire = logger.SERVER_CLOUDAPI_WIRE_LOGGER
        _sysadmin_cloudapi_client = \
            vcd_utils.get_cloudapi_client_from_vcd_client(
                _sysadmin_client,
                logger.SERVER_LOGGER,
                logger_wire)

        _sysadmin_user_context = user_context.UserContext(
            client=_sysadmin_client,
            cloud_api_client=_sysadmin_cloudapi_client)
        self._sysadmin_user_context_map[api_version] = _sysadmin_user_context

    def end(self):
        for api_version, sysadmin_user_context in \
                self._sysadmin_user_context_map.items():
            try:
                sysadmin_user_context.client.logout()
            except Exception:
                if not api_version:
                    api_version = "Default api version"
                msg = f"Failed to logout user: {sysadmin_user_context.name}" \
                    f"at api_version: {api_version}."
                logger.SERVER_LOGGER.debug(msg)
        self._user_context_map.clear()
        self._sysadmin_user_context_map.clear()



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause
import click
from vcd_cli.utils import stderr
from vcd_cli.utils import stdout

from container_service_extension.client.template import Template
import container_service_extension.client.utils as client_utils
from container_service_extension.logging.logger import CLIENT_LOGGER


@click.group(name='template',
             short_help='Manage native kubernetes runtime templates')
@click.pass_context
def template_group(ctx):
    """Manage native kubernetes runtime templates.

Templates that are no longer compliant with CSE template cookbook
specification will be ignored.

\b
Examples
    vcd cse template list
        Display templates that can be used to deploy native clusters.
    """
    pass


@template_group.command('list',
                        short_help='List native kubernetes runtime templates')
@click.pass_context
def list_templates(ctx):
    """Display templates that can be used to deploy native clusters."""
    CLIENT_LOGGER.debug(f'Executing command: {ctx.command_path}')
    try:
        client_utils.cse_restore_session(ctx)
        client = ctx.obj['client']
        template = Template(client)
        result = template.get_templates()
        CLIENT_LOGGER.debug(result)
        value_field_to_display_field = {
            'name': 'Name',
            'revision': 'Revision',
            'is_default': 'Default',
            'catalog': 'Catalog',
            'catalog_item': 'Catalog Item',
            'description': 'Description'
        }
        filtered_result = client_utils.filter_columns(result, value_field_to_display_field)  # noqa: E501
        stdout(filtered_result, ctx, sort_headers=False)
    except Exception as e:
        stderr(e, ctx)
        CLIENT_LOGGER.error(str(e))



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import os

import pyvcloud.vcd.exceptions as vcd_exceptions
import semantic_version
import yaml

import container_service_extension.client.constants as cli_constants
from container_service_extension.client.cse_client.api_35.native_cluster_api import NativeClusterApi  # noqa: E501
import container_service_extension.client.utils as client_utils
import container_service_extension.common.constants.shared_constants as shared_constants  # noqa: E501
import container_service_extension.common.utils.pyvcloud_utils as vcd_utils
import container_service_extension.exception.exceptions as cse_exceptions
import container_service_extension.logging.logger as logger
import container_service_extension.rde.common.entity_service as def_entity_svc
import container_service_extension.rde.constants as rde_constants
import container_service_extension.rde.models.common_models as common_models
import container_service_extension.rde.models.rde_1_0_0 as rde_1_0_0
import container_service_extension.rde.schema_service as def_schema_svc


class DEClusterNative:
    """Handle operations that are specific to cluster kind 'native'.

    Examples:
        cluster apply
        cluster create where cluster kind specified as CLI param
        cluster resize where cluster kind specified as CLI param

    """

    def __init__(self, client):
        logger_wire = logger.NULL_LOGGER
        if os.getenv(cli_constants.ENV_CSE_CLIENT_WIRE_LOGGING):
            logger_wire = logger.CLIENT_WIRE_LOGGER
        self._cloudapi_client = \
            vcd_utils.get_cloudapi_client_from_vcd_client(
                client=client, logger_debug=logger.CLIENT_LOGGER,
                logger_wire=logger_wire)
        self._native_cluster_api = NativeClusterApi(client)
        self._client = client
        schema_service = def_schema_svc.DefSchemaService(self._cloudapi_client)
        self._server_rde_version = \
            schema_service.get_latest_registered_schema_version()

    def create_cluster(self, cluster_entity: rde_1_0_0.NativeEntity):
        """Create a new Kubernetes cluster.

        :param models.NativeEntity cluster_entity: native cluster entity
        :return: (json) A parsed json object describing the requested cluster.
        """
        msg = "Operation not supported; Under implementation"
        raise vcd_exceptions.OperationNotSupportedException(msg)

    def resize_cluster(self, cluster_entity: rde_1_0_0.NativeEntity):
        """Resize the existing Kubernetes cluster.

        :param models.NativeEntity cluster_entity: native cluster entity
        :return: (json) A parsed json object describing the requested cluster.
        """
        msg = "Operation not supported; Under implementation"
        raise vcd_exceptions.OperationNotSupportedException(msg)

    def get_cluster_info(self, cluster_name, cluster_id=None,
                         org=None, vdc=None, **kwargs):
        """Get cluster information using DEF API.

        :param str cluster_name: name of the cluster
        :param str vdc: name of vdc
        :param str org: name of org
        :param kwargs: *filter (dict): keys,values for DEF API query filter

        :return: cluster information
        :rtype: dict
        """
        if cluster_id:
            return self.get_cluster_info_by_id(cluster_id)
        filters = client_utils.construct_filters(
            self._server_rde_version, org=org, vdc=vdc)
        entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
        def_entity = \
            entity_svc.get_native_rde_by_name_and_rde_version(
                cluster_name, self._server_rde_version, filters=filters)  # noqa: E501
        logger.CLIENT_LOGGER.debug(f"Defined entity info from server:{def_entity}")  # noqa: E501
        if not def_entity:
            logger.CLIENT_LOGGER.error(f"Cannot find native cluster with name {cluster_name}")  # noqa: E501
            raise cse_exceptions.ClusterNotFoundError(f"Cluster '{cluster_name}' not found.")  # noqa: E501
        # TODO() relevant output
        if def_entity:
            return yaml.dump(def_entity.entity.to_dict())

    def get_cluster_info_by_id(self, cluster_id, **kwargs):
        """Get cluster information by cluster ID.

        :param str cluster_id: ID of the cluster
        :return: cluster information in yaml format
        :rtype: str
        """
        entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
        def_entity = entity_svc.get_entity(cluster_id)
        logger.CLIENT_LOGGER.debug(f"Defined entity info from server: {def_entity}")  # noqa: E501
        return yaml.dump(def_entity.entity.to_dict())

    def delete_cluster(self, cluster_name, cluster_id=None,
                       org=None, vdc=None):
        """Delete DEF native cluster by name.

        :param str cluster_name: native cluster name
        :param str org: name of the org
        :param str vdc: name of the vdc
        :return: string containing delete operation task href
        :rtype: str
        :raises ClusterNotFoundError
        """
        if not cluster_id:
            cluster_id = self.get_cluster_id_by_name(cluster_name, org, vdc)
        return self.delete_cluster_by_id(cluster_id)

    def delete_cluster_by_id(self, cluster_id, **kwargs):
        """Delete the existing Kubernetes cluster by id.

        :param str cluster_id: native cluster entity id
        :return: string containing the task for delete operation
        :rtype: str
        """
        cluster_entity = \
            self._native_cluster_api.delete_cluster_by_cluster_id(cluster_id)
        task_href = cluster_entity.entity.status.task_href
        return client_utils.construct_task_console_message(task_href)  # noqa: E501

    def delete_nfs_node(self, cluster_name, node_name, org=None, vdc=None):
        """Delete nfs node given the cluster name and node name.

        :param str cluster_name: native cluster name
        :param str node_name: nfs-node name
        :param str org: name of the org
        :param str vdc: name of the vdc
        :return: string containing delete operation task href
        :rtype: str
        :raises ClusterNotFoundError
        """
        filters = client_utils.construct_filters(
            self._server_rde_version, org=org, vdc=vdc)
        entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
        def_entity = entity_svc.get_native_rde_by_name_and_rde_version(
            cluster_name, self._server_rde_version, filters=filters)
        if def_entity:
            return self.delete_nfs_by_cluster_id(def_entity.id, node_name)
        raise cse_exceptions.ClusterNotFoundError(f"Cluster '{cluster_name}' not found.")  # noqa: E501

    def delete_nfs_by_cluster_id(self, cluster_id, node_name):
        """Delete the nfs-node by name from the given cluster id.

        :param str cluster_id: native cluster entity id
        :param str node_name: nfs-node name
        :return: string containing the task for delete operation
        :rtype: str
        """
        cluster_entity = \
            self._native_cluster_api.delete_nfs_node_by_node_name(cluster_id,
                                                                  node_name)
        task_href = cluster_entity.entity.status.task_href
        return client_utils.construct_task_console_message(task_href)

    def get_cluster_config(self, cluster_name, cluster_id=None,
                           org=None, vdc=None):
        """Get cluster config for the given cluster name.

        :param str cluster_name: name of the cluster
        :param str cluster_id: id of the cluster
        :param str vdc: name of vdc
        :param str org: name of org

        :return: cluster information
        :rtype: str
        :raises ClusterNotFoundError, CseDuplicateClusterError
        """
        if not cluster_id:
            cluster_id = self.get_cluster_id_by_name(cluster_name, org, vdc)
        return self.get_cluster_config_by_id(cluster_id)

    def get_cluster_config_by_id(self, cluster_id, **kwargs):
        """Get the cluster config for given cluster id.

        :param str cluster_id: native cluster entity id
        :return: decoded response content
        :rtype: dict
        """
        return self._native_cluster_api.get_cluster_config_by_cluster_id(cluster_id)  # noqa: E501

    def get_upgrade_plan(self, cluster_name, org=None, vdc=None):
        """Get the upgrade plan for given cluster.

        :param cluster_name: name of the cluster
        :param org: name of the org
        :param vdc: name of the vdc
        :return: upgrade plan info
        :rtype: dict
        """
        filters = client_utils.construct_filters(
            self._server_rde_version, org=org, vdc=vdc)
        entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
        def_entity = entity_svc.get_native_rde_by_name_and_rde_version(
            cluster_name, self._server_rde_version, filters=filters)
        if def_entity:
            return self.get_upgrade_plan_by_cluster_id(def_entity.id)
        raise cse_exceptions.ClusterNotFoundError(f"Cluster '{cluster_name}' not found.")  # noqa: E501

    def get_upgrade_plan_by_cluster_id(self, cluster_id, **kwargs):
        """Get the upgrade plan for give cluster id.

        :param cluster_id: unique id of the cluster
        :return: decoded response content
        :rtype: list
        """
        return self._native_cluster_api.get_upgrade_plan_by_cluster_id(cluster_id)  # noqa: E501

    def upgrade_cluster(self, cluster_name, template_name,
                        template_revision, org_name=None, ovdc_name=None):
        """Get the upgrade plan for given cluster.

        :param str cluster_name: name of the cluster
        :param str template_name: Name of the template the cluster should be
        upgraded to.
        :param str template_revision: Revision of the template the cluster
        should be upgraded to.
        :param org_name: name of the org
        :param ovdc_name: name of the vdc
        :return: string containing upgrade cluster task href
        :rtype: str
        """
        filters = client_utils.construct_filters(
            self._server_rde_version, org=org_name, vdc=ovdc_name)
        entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
        current_entity = entity_svc.get_native_rde_by_name_and_rde_version(
            cluster_name, self._server_rde_version, filters=filters)
        # NOTE: This function is only valid when the CSE server is running
        # with RDE 1.0.0
        if current_entity:
            current_entity.entity.spec.k8_distribution.template_name = template_name  # noqa: E501
            current_entity.entity.spec.k8_distribution.template_revision = template_revision  # noqa: E501
            return self.upgrade_cluster_by_cluster_id(current_entity.id, cluster_def_entity=current_entity.to_dict())  # noqa: E501
        raise cse_exceptions.ClusterNotFoundError(f"Cluster '{cluster_name}' not found.")  # noqa: E501

    def upgrade_cluster_by_cluster_id(self, cluster_id, cluster_def_entity, **kwargs):  # noqa: E501
        """Get the upgrade plan for give cluster id.

        :param str cluster_id: unique id of the cluster
        :param dict cluster_def_entity: defined entity
        :return: string containing upgrade cluster task href
        :rtype: str
        """
        # TODO: check if we really need to decode-encode-decode-encode
        cluster_upgrade_definition = common_models.DefEntity(**cluster_def_entity)  # noqa: E501
        cluster_def_entity = \
            self._native_cluster_api.upgrade_cluster_by_cluster_id(
                cluster_id, cluster_upgrade_definition)
        task_href = cluster_def_entity.entity.status.task_href
        return client_utils.construct_task_console_message(task_href)

    def _get_cluster_name_from_cluster_apply_specification(self, input_spec: dict):  # noqa: E501
        """Derive cluster name from cluster apply specificaiton.

        :param dict input_spec: Input specification
        :return: cluster name
        :rtype: str
        """
        # RDE version > 1.0.0 will have the cluster name under metadata.name
        metadata: dict = input_spec.get('metadata')
        if semantic_version.Version(self._server_rde_version) <= \
                semantic_version.Version(rde_constants.RDEVersion.RDE_1_0_0):
            return metadata.get('cluster_name')
        return metadata.get('name')

    def apply(self, cluster_apply_spec: dict, cluster_id: str = None, **kwargs):  # noqa: E501
        """Apply the configuration either to create or update the cluster.

        :param dict cluster_config: cluster configuration information
        :return: dictionary containing the apply operation task
        :rtype: dict
        """
        cluster_name = \
            self._get_cluster_name_from_cluster_apply_specification(cluster_apply_spec)  # noqa: E501
        # cluster name should not be missing from the apply specification
        if not cluster_name:
            raise Exception('Cluster name missing in the cluster apply specification')  # noqa: E501
        entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
        if cluster_id:
            # If cluster id doesn't exist, an exception will be raised
            def_entity = entity_svc.get_entity(cluster_id)
        else:
            def_entity = entity_svc.get_native_rde_by_name_and_rde_version(
                cluster_name, self._server_rde_version)
        if not def_entity:
            cluster_def_entity = self._native_cluster_api.create_cluster(
                cluster_apply_spec)
        else:
            cluster_id = def_entity.id
            cluster_def_entity = \
                self._native_cluster_api.update_cluster_by_cluster_id(
                    cluster_id, cluster_apply_spec)
        task_href = cluster_def_entity.entity.status.task_href
        return client_utils.construct_task_console_message(task_href)

    def get_cluster_id_by_name(self, cluster_name, org=None, vdc=None):
        filters = client_utils.construct_filters(
            self._server_rde_version, org=org, vdc=vdc)
        entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
        def_entity = entity_svc.get_native_rde_by_name_and_rde_version(
            cluster_name, self._server_rde_version, filters=filters)
        if not def_entity:
            raise cse_exceptions.ClusterNotFoundError(f"Cluster '{cluster_name}' not found.")  # noqa: E501
        return def_entity.id

    def share_cluster(self, cluster_id, cluster_name, users: list,
                      access_level_id, org, vdc):
        """Share cluster with passed in users."""
        if not cluster_id:
            cluster_id = self.get_cluster_id_by_name(cluster_name, org, vdc)
        org_href = self._client.get_org_by_name(org).get('href')
        name_to_id: dict = client_utils.create_user_name_to_id_dict(
            self._client, users, org_href)

        # Parse user id info
        update_acl_entries = []
        for username, user_id in name_to_id.items():
            acl_entry = common_models.ClusterAclEntry(
                memberId=user_id,
                username=username,
                accessLevelId=access_level_id)
            update_acl_entries.append(acl_entry)

        # Only retain entries that are not updated
        for acl_entry in self._native_cluster_api.\
                list_native_cluster_acl_entries(cluster_id):
            username = acl_entry.username
            if name_to_id.get(username):
                # Check that access level is not reduced
                curr_access_level_id = acl_entry.accessLevelId
                if client_utils.access_level_reduced(
                        access_level_id, curr_access_level_id):
                    raise Exception(f'{username} currently has higher access '
                                    f'level: {curr_access_level_id}')
            else:
                update_acl_entries.append(acl_entry)

        update_acl_values = \
            [acl_entry.construct_filtered_dict(include=cli_constants.CLUSTER_ACL_UPDATE_REQUEST_FIELDS)  # noqa: E501
             for acl_entry in update_acl_entries]
        self._native_cluster_api.put_cluster_acl(cluster_id, update_acl_values)

    def unshare_cluster(self, cluster_id, cluster_name, users: list, org=None,
                        vdc=None):
        if not cluster_id:
            cluster_id = self.get_cluster_id_by_name(cluster_name, org, vdc)

        delete_users_set = set(users)
        updated_acl_entries = []
        for acl_entry in self._native_cluster_api.\
                list_native_cluster_acl_entries(cluster_id):
            if acl_entry.username not in delete_users_set:
                acl_dict = acl_entry.construct_filtered_dict(
                    include=cli_constants.CLUSTER_ACL_UPDATE_REQUEST_FIELDS)
                updated_acl_entries.append(acl_dict)
            else:
                delete_users_set.remove(acl_entry.username)

        if len(delete_users_set) > 0:
            raise Exception(f'Cluster {cluster_name or cluster_id} is not '
                            f'currently shared with: {list(delete_users_set)}')

        self._native_cluster_api.put_cluster_acl(cluster_id, updated_acl_entries)  # noqa: E501

    def list_share_entries(self, cluster_id, cluster_name, org=None, vdc=None):
        if not cluster_id:
            cluster_id = self.get_cluster_id_by_name(cluster_name, org, vdc)
        result_count = page_num = 0
        while True:
            page_num += 1
            response_body = self._native_cluster_api.get_single_page_cluster_acl(  # noqa: E501
                cluster_id=cluster_id,
                page=page_num,
                page_size=cli_constants.CLI_ENTRIES_PER_PAGE)
            result_total = response_body[shared_constants.PaginationKey.RESULT_TOTAL]  # noqa: E501
            acl_values = response_body[shared_constants.PaginationKey.VALUES]
            if not acl_values:
                break
            result_count += len(acl_values)
            yield acl_values, result_count < result_total



# coding: utf-8

"""
    PKS

    PKS API  # noqa: E501

    OpenAPI spec version: 1.1.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six


class TaskInfo(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'uuid': 'str',
        'type': 'str',
        'status': 'str',
        'clusters': 'list[ClusterUpgradeInfo]',
        'start_time': 'int',
        'end_time': 'int'
    }

    attribute_map = {
        'uuid': 'uuid',
        'type': 'type',
        'status': 'status',
        'clusters': 'clusters',
        'start_time': 'start_time',
        'end_time': 'end_time'
    }

    def __init__(self, uuid=None, type=None, status=None, clusters=None, start_time=None, end_time=None):  # noqa: E501
        """TaskInfo - a model defined in Swagger"""  # noqa: E501

        self._uuid = None
        self._type = None
        self._status = None
        self._clusters = None
        self._start_time = None
        self._end_time = None
        self.discriminator = None

        if uuid is not None:
            self.uuid = uuid
        if type is not None:
            self.type = type
        if status is not None:
            self.status = status
        if clusters is not None:
            self.clusters = clusters
        if start_time is not None:
            self.start_time = start_time
        if end_time is not None:
            self.end_time = end_time

    @property
    def uuid(self):
        """Gets the uuid of this TaskInfo.  # noqa: E501

        The uuid of a task  # noqa: E501

        :return: The uuid of this TaskInfo.  # noqa: E501
        :rtype: str
        """
        return self._uuid

    @uuid.setter
    def uuid(self, uuid):
        """Sets the uuid of this TaskInfo.

        The uuid of a task  # noqa: E501

        :param uuid: The uuid of this TaskInfo.  # noqa: E501
        :type: str
        """

        self._uuid = uuid

    @property
    def type(self):
        """Gets the type of this TaskInfo.  # noqa: E501

        The type of a task  # noqa: E501

        :return: The type of this TaskInfo.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this TaskInfo.

        The type of a task  # noqa: E501

        :param type: The type of this TaskInfo.  # noqa: E501
        :type: str
        """
        allowed_values = ["UPGRADE"]  # noqa: E501
        if type not in allowed_values:
            raise ValueError(
                "Invalid value for `type` ({0}), must be one of {1}"  # noqa: E501
                .format(type, allowed_values)
            )

        self._type = type

    @property
    def status(self):
        """Gets the status of this TaskInfo.  # noqa: E501


        :return: The status of this TaskInfo.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this TaskInfo.


        :param status: The status of this TaskInfo.  # noqa: E501
        :type: str
        """
        allowed_values = ["not started", "in progress", "canceled", "done"]  # noqa: E501
        if status not in allowed_values:
            raise ValueError(
                "Invalid value for `status` ({0}), must be one of {1}"  # noqa: E501
                .format(status, allowed_values)
            )

        self._status = status

    @property
    def clusters(self):
        """Gets the clusters of this TaskInfo.  # noqa: E501


        :return: The clusters of this TaskInfo.  # noqa: E501
        :rtype: list[ClusterUpgradeInfo]
        """
        return self._clusters

    @clusters.setter
    def clusters(self, clusters):
        """Sets the clusters of this TaskInfo.


        :param clusters: The clusters of this TaskInfo.  # noqa: E501
        :type: list[ClusterUpgradeInfo]
        """

        self._clusters = clusters

    @property
    def start_time(self):
        """Gets the start_time of this TaskInfo.  # noqa: E501


        :return: The start_time of this TaskInfo.  # noqa: E501
        :rtype: int
        """
        return self._start_time

    @start_time.setter
    def start_time(self, start_time):
        """Sets the start_time of this TaskInfo.


        :param start_time: The start_time of this TaskInfo.  # noqa: E501
        :type: int
        """

        self._start_time = start_time

    @property
    def end_time(self):
        """Gets the end_time of this TaskInfo.  # noqa: E501


        :return: The end_time of this TaskInfo.  # noqa: E501
        :rtype: int
        """
        return self._end_time

    @end_time.setter
    def end_time(self, end_time):
        """Sets the end_time of this TaskInfo.


        :param end_time: The end_time of this TaskInfo.  # noqa: E501
        :type: int
        """

        self._end_time = end_time

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(TaskInfo, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, TaskInfo):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import threading

from container_service_extension.common.utils.init_utils import run_once


# Thread data to be initialized once. This data is specific to each thread.
THREAD_DATA = None


@run_once
def init_thread_local_data():
    """Initialize THREAD_DATA once for all threads.

    Running this function more than once creates a new instance of
    threading.local() each time, overwriting the original THREAD_DATA. The
    value of THREAD_DATA will look different to each thread.
    """
    global THREAD_DATA
    THREAD_DATA = threading.local()


def set_thread_local_data(name: str, value):
    """Set the attribute of current thread to the given value."""
    global THREAD_DATA
    setattr(THREAD_DATA, name, value)


def get_thread_local_data(name: str):
    """Get the value of given attribute from the current thread."""
    global THREAD_DATA
    return getattr(THREAD_DATA, name, None)


def get_thread_local_data_as_dict() -> dict:
    """Get all the current thread attributes as dictionary."""
    global THREAD_DATA
    return THREAD_DATA.__dict__


def set_thread_local_data_from_dict(data_dict: dict = {}):
    """Set the current thread attributes to given values found in dictionary.

    :param dict data_dict: input attributes
    :return:
    """
    global THREAD_DATA
    for key in data_dict:
        THREAD_DATA.__dict__[key] = data_dict.get(key)


def reset_thread_local_data():
    """Reset the current thread attributes.

    :param dict data_dict: input attributes
    :return:
    """
    global THREAD_DATA
    THREAD_DATA.__dict__.clear()



# container-service-extension
# Copyright (c) 2019 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""Constants used only in the CSE server."""

from dataclasses import dataclass
from enum import Enum
from enum import unique

import requests
import semantic_version

# CSE SERVICE;
# keys for CSE info on the extension
CSE_VERSION_KEY = 'cse_version'
LEGACY_MODE_KEY = 'legacy_mode'
RDE_VERSION_IN_USE_KEY = 'rde_version'
# default version strings for CSE api extension
UNKNOWN_CSE_VERSION = semantic_version.Version("0.0.0")
UNKNOWN_VCD_API_VERSION = "0.0"

# CSE SERVICE; used for registering CSE to vCD as an api extension service.
CSE_SERVICE_NAME = 'cse'
CSE_SERVICE_NAMESPACE = 'cse'
EXCHANGE_TYPE = 'direct'

# DEPLOY RIGHTS; used by authorization framework to weed out unauthorized calls
CSE_NATIVE_DEPLOY_RIGHT_NAME = 'CSE NATIVE DEPLOY RIGHT'
CSE_NATIVE_DEPLOY_RIGHT_DESCRIPTION = 'Right necessary to deploy kubernetes ' \
    'cluster via vCD apis in CSE'
CSE_NATIVE_DEPLOY_RIGHT_CATEGORY = 'cseRight'
CSE_NATIVE_DEPLOY_RIGHT_BUNDLE_KEY = 'cseBundleKey'
CSE_PKS_DEPLOY_RIGHT_NAME = 'PKS DEPLOY RIGHT'
CSE_PKS_DEPLOY_RIGHT_DESCRIPTION = 'Right necessary to deploy kubernetes ' \
    'cluster via vSphere apis in PKS'
CSE_PKS_DEPLOY_RIGHT_CATEGORY = 'pksRight'
CSE_PKS_DEPLOY_RIGHT_BUNDLE_KEY = 'pksBundleKey'

# KUBERNETES PROVIDERS; used by server operations related to k8s providers
K8S_PROVIDER_KEY = 'k8s_provider'
PKS_PLANS_KEY = 'pks_plans'
PKS_CLUSTER_DOMAIN_KEY = 'pks_cluster_domain'
PKS_COMPUTE_PROFILE_KEY = 'pks_compute_profile_name'

# PKS API endpoint version
VERSION_V1 = 'v1'

# CSE global pvdc compute policy name
CSE_GLOBAL_PVDC_COMPUTE_POLICY_NAME = 'global'
CSE_GLOBAL_PVDC_COMPUTE_POLICY_DESCRIPTION = \
    'global PVDC compute policy for cse'

# CSE cluster Kube Config path
CSE_CLUSTER_KUBECONFIG_PATH = '/root/.kube/config'

# MQTT constants
ADMIN_EXT_SERVICE_PATH = 'admin/extension/service'
API_FILTER_PATH = 'apifilter'
API_FILTERS_PATH = 'apifilters'
EXTENSIONS_API_PATH = 'extensions/api'
MQTT_API_FILTER_PATTERN = '/api/mqttEndpoint/cse'  # Needs to start with '/'
MQTT_EXTENSION_VERSION = '1.0.0'
MQTT_EXTENSION_VENDOR = 'VMWare'
MQTT_EXTENSION_URN = 'urn:vcloud:extension-api:' + MQTT_EXTENSION_VENDOR + ':'\
                     + CSE_SERVICE_NAME + ':' + MQTT_EXTENSION_VERSION
MQTT_EXTENSION_PRIORITY = 100
MQTT_MIN_API_VERSION = 35.0
MQTT_TOKEN_NAME = "mqttCseToken"
TOKEN_PATH = 'tokens'


# Encryption constants
PBKDF2_ITERATIONS = 100000
SALT_SIZE = 32
PBKDF2_OUTPUT_SIZE = 32

# Names of Message Consumer Thread and Watchdog Thread
MESSAGE_CONSUMER_THREAD = 'MessageConsumer'
WATCHDOG_THREAD = 'ConsumerWatchdog'


# Config file error messages
CONFIG_DECRYPTION_ERROR_MSG = \
    "Config file decryption failed: invalid decryption password"
VCENTER_LOGIN_ERROR_MSG = "vCenter login failed (check config file for "\
    "vCenter username/password)."

# Request Id format for logging
REQUEST_ID_FORMAT = 'Request Id: %(requestId)s | '


@unique
class OperationType(str, Enum):
    NATIVE_CLUSTER = 'nativecluster'
    CLUSTER = 'cluster'
    NODE = 'node'
    OVDC = 'ovdc'
    SYSTEM = 'system'
    TEMPLATE = 'template'
    ORG_VDCS = 'orgvdc'


@unique
class ThreadLocalData(str, Enum):
    USER_AGENT = 'User-Agent'
    REQUEST_ID = 'request_id'


@unique
class DefEntityOperation(str, Enum):
    CREATE = 'CREATE'
    DELETE = 'DELETE'
    UPDATE = 'UPDATE'
    UPGRADE = 'UPGRADE'
    UNKNOWN = 'UNKNOWN'


@unique
class DefEntityOperationStatus(str, Enum):
    IN_PROGRESS = 'IN_PROGRESS'
    SUCCEEDED = 'SUCCEEDED'
    FAILED = 'FAILED'
    UNKNOWN = 'UNKNOWN'


@unique
class FlattenedClusterSpecKey1X(Enum):
    WORKERS_COUNT = 'workers.count'
    NFS_COUNT = 'nfs.count'
    TEMPLATE_NAME = 'k8_distribution.template_name'
    TEMPLATE_REVISION = 'k8_distribution.template_revision'
    EXPOSE = 'expose'


@unique
class FlattenedClusterSpecKey2X(Enum):
    WORKERS_COUNT = 'topology.workers.count'
    WORKERS_SIZING_CLASS = 'topology.workers.sizingClass'
    WORKERS_STORAGE_PROFILE = 'topology.workers.storageProfile'
    NFS_COUNT = 'topology.nfs.count'
    NFS_SIZING_CLASS = 'topology.nfs.sizingClass'
    NFS_STORAGE_PROFILE = 'topology.nfs.storageProfile'
    TEMPLATE_NAME = 'distribution.templateName'
    TEMPLATE_REVISION = 'distribution.templateRevision'
    EXPOSE = 'settings.network.expose'


VALID_UPDATE_FIELDS_2X = \
    [FlattenedClusterSpecKey2X.WORKERS_COUNT.value,
     FlattenedClusterSpecKey2X.NFS_COUNT.value,
     FlattenedClusterSpecKey2X.TEMPLATE_NAME.value,
     FlattenedClusterSpecKey2X.TEMPLATE_REVISION.value,
     FlattenedClusterSpecKey2X.EXPOSE.value]


CLUSTER_ENTITY = 'cluster_entity'


# User ID parsing
USER_PATH = '/user/'
ADMIN_USER_PATH = '/admin/user/'

# vApp Access control type
CHANGE_ACCESS = 'Change'


@unique
class NodeType(str, Enum):
    CONTROL_PLANE = 'mstr'
    WORKER = 'node'
    NFS = 'nfsd'


@unique
class K8sProvider(str, Enum):
    """Types of Kubernetes providers.

    Having a str mixin allows us to do things like:
    'native' == K8sProvider.NATIVE
    f"Kubernetes provider is '{K8sProvider.NATIVE}'
    """

    NATIVE = 'native'
    PKS = 'ent-pks'
    NONE = 'none'


@unique
class KwargKey(str, Enum):
    """Types of keyword arguments for cluster brokers."""

    DATA = 'data'
    TELEMETRY = 'telemetry'


@unique
class ClusterScriptFile(str, Enum):
    """Types of script for vApp template customizations in CSE."""

    SCRIPTS_DIR = "cluster_scripts"
    CONTROL_PLANE = 'control_plane.sh'
    NODE = 'node.sh'

    # Note that we need _ in the version instead of dots to allow
    # python package parsing methods to work well
    VERSION_1_X = "v1_x"
    VERSION_2_X = "v2_x"


@unique
class TemplateScriptFile(str, Enum):
    """Types of script for vApp template customizations in CSE."""

    # vapp template creation scripts
    CUST = 'cust.sh'
    INIT = 'init.sh'
    NFSD = 'nfsd.sh'

    # cluster upgrade scripts
    DOCKER_UPGRADE = 'cluster-upgrade/docker-upgrade.sh'
    CONTROL_PLANE_CNI_APPLY = 'cluster-upgrade/control-plane-cni-apply.sh'
    CONTROL_PLANE_K8S_UPGRADE = 'cluster-upgrade/control-plane-k8s-upgrade.sh'
    WORKER_K8S_UPGRADE = 'cluster-upgrade/worker-k8s-upgrade.sh'


@unique
class ScriptFile(str, Enum):
    # For Usage in vcdbroker.py due to legacy code.
    # When those are deprecated, we could remove this struct.
    """Types of script for vApp template customizations in CSE."""

    # vapp template creation scripts
    CUST = 'cust.sh'
    INIT = 'init.sh'
    NFSD = 'nfsd.sh'

    # cluster initialization scripts
    CONTROL_PLANE = 'mstr.sh'
    NODE = 'node.sh'

    # cluster upgrade scripts
    DOCKER_UPGRADE = 'cluster-upgrade/docker-upgrade.sh'
    CONTROL_PLANE_CNI_APPLY = 'cluster-upgrade/master-cni-apply.sh'
    CONTROL_PLANE_K8S_UPGRADE = 'cluster-upgrade/master-k8s-upgrade.sh'
    WORKER_K8S_UPGRADE = 'cluster-upgrade/worker-k8s-upgrade.sh'


@unique
class LegacyLocalTemplateKey(str, Enum):
    """Enumerate the keys that define a legacy template.

    All the keys for a template except min_cse_version and
    max_cse_version.
    """

    CATALOG_ITEM_NAME = 'catalog_item_name'
    CNI = 'cni'
    CNI_VERSION = 'cni_version'
    COMPUTE_POLICY = 'compute_policy'
    CPU = 'cpu'
    DEPRECATED = 'deprecated'
    DESCRIPTION = 'description'
    DOCKER_VERSION = 'docker_version'
    KIND = 'kind'
    KUBERNETES = 'kubernetes'
    KUBERNETES_VERSION = 'kubernetes_version'
    MEMORY = 'mem'
    NAME = 'name'
    OS = 'os'
    REVISION = 'revision'
    UPGRADE_FROM = 'upgrade_from'


@unique
class LocalTemplateKey(str, Enum):
    """Enumerate the keys that define a template."""

    CATALOG_ITEM_NAME = 'catalog_item_name'
    CNI = 'cni'
    CNI_VERSION = 'cni_version'
    CPU = 'cpu'
    DEPRECATED = 'deprecated'
    DESCRIPTION = 'description'
    DOCKER_VERSION = 'docker_version'
    KIND = 'kind'
    KUBERNETES = 'kubernetes'
    KUBERNETES_VERSION = 'kubernetes_version'
    MEMORY = 'mem'
    NAME = 'name'
    OS = 'os'
    REVISION = 'revision'
    COOKBOOK_VERSION = 'cookbook_version'
    UPGRADE_FROM = 'upgrade_from'
    MIN_CSE_VERSION = 'min_cse_version'
    MAX_CSE_VERSION = 'max_cse_version'


@unique
class RemoteTemplateKeyV1(str, Enum):
    """Enumerate the keys that define a template."""

    CNI = 'cni'
    CNI_VERSION = 'cni_version'
    COMPUTE_POLICY = 'compute_policy'
    CPU = 'cpu'
    DEPRECATED = 'deprecated'
    DESCRIPTION = 'description'
    DOCKER_VERSION = 'docker_version'
    KIND = 'kind'
    KUBERNETES = 'kubernetes'
    KUBERNETES_VERSION = 'kubernetes_version'
    MEMORY = 'mem'
    NAME = 'name'
    OS = 'os'
    REVISION = 'revision'
    SOURCE_OVA_HREF = 'source_ova'
    SOURCE_OVA_NAME = 'source_ova_name'
    SOURCE_OVA_SHA256 = 'sha256_ova'
    UPGRADE_FROM = 'upgrade_from'


@unique
class RemoteTemplateKeyV2(str, Enum):
    """Enumerate the keys that define a template."""

    CNI = 'cni'
    CNI_VERSION = 'cni_version'
    CPU = 'cpu'
    DEPRECATED = 'deprecated'
    DESCRIPTION = 'description'
    DOCKER_VERSION = 'docker_version'
    KIND = 'kind'
    KUBERNETES = 'kubernetes'
    KUBERNETES_VERSION = 'kubernetes_version'
    MEMORY = 'mem'
    NAME = 'name'
    OS = 'os'
    REVISION = 'revision'
    SOURCE_OVA_HREF = 'source_ova'
    SOURCE_OVA_NAME = 'source_ova_name'
    SOURCE_OVA_SHA256 = 'sha256_ova'
    UPGRADE_FROM = 'upgrade_from'
    MIN_CSE_VERSION = 'min_cse_version'
    MAX_CSE_VERSION = 'max_cse_version'


@unique
class RemoteTemplateCookbookVersion(Enum):
    """Enumerate the remote template cookbook versions."""

    Version1 = semantic_version.Version('1.0.0')
    Version2 = semantic_version.Version('2.0.0')


# CSE requests
@unique
class CseOperation(Enum):
    def __init__(self, description, api_path_format, ideal_response_code=requests.codes.ok):  # noqa: E501
        self._description = description
        self._ideal_response_code = ideal_response_code
        self._api_path_format = api_path_format

    @property
    def ideal_response_code(self):
        return int(self._ideal_response_code)

    @property
    def api_path_format(self):
        return self._api_path_format

    CLUSTER_CONFIG = ('get config of cluster', '/cse/cluster/%s/config')
    CLUSTER_CREATE = ('create cluster', '/cse/clusters', requests.codes.accepted)  # noqa: E501
    CLUSTER_DELETE = ('delete cluster', '/cse/cluster/%s', requests.codes.accepted)  # noqa: E501
    CLUSTER_INFO = ('get info of cluster', '/cse/cluster/%s')
    NATIVE_CLUSTER_LIST = ('list legacy clusters', '/cse/nativeclusters')
    CLUSTER_LIST = ('list clusters', '/cse/clusters')
    CLUSTER_RESIZE = ('resize cluster', '/cse/cluster/%s', requests.codes.accepted)  # noqa: E501
    CLUSTER_UPGRADE_PLAN = ('get supported cluster upgrade paths', '/cse/cluster/%s/upgrade-plan')  # noqa: E501
    CLUSTER_UPGRADE = ('upgrade cluster software', '/cse/cluster/%s/action/upgrade', requests.codes.accepted)  # noqa: E501
    NODE_CREATE = ('create node', '/cse/nodes', requests.codes.accepted)
    NODE_DELETE = ('delete node', '/cse/nodes/%s', requests.codes.accepted)
    NODE_INFO = ('get info of node', '/cse/nodes/%s')

    V35_CLUSTER_CONFIG = ('get config of DEF cluster', '/cse/3.0/cluster/%s/config')  # noqa: E501
    V35_CLUSTER_CREATE = ('create DEF cluster', '/cse/3.0/clusters', requests.codes.accepted)  # noqa: E501
    V35_CLUSTER_DELETE = ('delete DEF cluster', '/cse/3.0/cluster/%s', requests.codes.accepted)  # noqa: E501
    V35_CLUSTER_INFO = ('get info of DEF cluster', '/cse/3.0/cluster/%s')
    V35_NATIVE_CLUSTER_LIST = ('list paginated DEF clusters', '/cse/3.0/nativeclusters')  # noqa: E501
    V35_CLUSTER_LIST = ('list DEF clusters', '/cse/3.0/clusters')
    V35_CLUSTER_RESIZE = ('resize DEF cluster', '/cse/3.0/cluster/%s', requests.codes.accepted)  # noqa: E501
    V35_CLUSTER_UPGRADE_PLAN = ('get supported DEF cluster upgrade paths', '/cse/3.0/cluster/%s/upgrade-plan')  # noqa: E501
    V35_CLUSTER_UPGRADE = ('upgrade DEF cluster software', '/cse/3.0/cluster/%s/action/upgrade', requests.codes.accepted)  # noqa: E501
    V35_CLUSTER_ACL_LIST = ('list cluster acl', '/cse/3.0/cluster/%s/acl')
    V35_CLUSTER_ACL_UPDATE = ('update cluster acl', '/cse/3.0/cluster/%s/acl', requests.codes.no_content)  # noqa: E501
    V35_NODE_CREATE = ('create DEF node', 'NOT IMPLEMENTED', requests.codes.accepted)  # noqa: E501
    V35_NODE_DELETE = ('delete DEF node', '/cse/3.0/cluster/%s/nfs/%s', requests.codes.accepted)  # noqa: E501
    V35_NODE_INFO = ('get info of DEF node', 'NOT IMPLEMENTED')

    V36_CLUSTER_CREATE = ('create DEF v36 cluster', '/cse/3.0/clusters', requests.codes.accepted)  # noqa: E501
    V36_CLUSTER_LIST = ('list V36 DEF clusters', '/cse/3.0/clusters')  # noqa: E501
    V36_NATIVE_CLUSTER_LIST = ('list paginated V36 DEF clusters', '/cse/3.0/nativeclusters')  # noqa: E501
    V36_CLUSTER_INFO = ('get info of DEF V36 clusters', '/cse/3.0/cluster/%s')
    V36_CLUSTER_CONFIG = ('get config of DEF V36 cluster', '/cse/3.0/cluster/%s/config')  # noqa: E501
    V36_CLUSTER_UPDATE = ('update DEF V36 cluster', '/cse/3.0/cluster/%s', requests.codes.accepted)  # noqa: E501
    V36_CLUSTER_DELETE = ('delete DEF V36 cluster', '/cse/3.0/cluster/%s', requests.codes.accepted)  # noqa: E501
    V36_CLUSTER_UPGRADE_PLAN = ('get supported V36 DEF cluster upgrade paths', '/cse/3.0/cluster/%s/upgrade-plan')  # noqa: E501
    V36_NODE_DELETE = ('delete V36 DEF node', '/cse/3.0/cluster/%s/nfs/%s', requests.codes.accepted)  # noqa: E501
    V36_CLUSTER_ACL_LIST = ('list V36 cluster acl', '/cse/3.0/cluster/%s/acl')
    V36_CLUSTER_ACL_UPDATE = ('update V36 cluster acl', '/cse/3.0/cluster/%s/acl', requests.codes.no_content)  # noqa: E501

    OVDC_UPDATE = ('enable or disable ovdc for k8s', '/cse/ovdc/%s', requests.codes.accepted)  # noqa: E501
    OVDC_INFO = ('get info of ovdc', '/cse/ovdc/%s')
    OVDC_LIST = ('list ovdcs', '/cse/ovdcs')
    ORG_VDC_LIST = ('list org VDCs', '/cse/orgvdcs')
    OVDC_COMPUTE_POLICY_LIST = ('list ovdc compute policies', '/cse/ovdc/%s/compute-policies')  # noqa: E501
    OVDC_COMPUTE_POLICY_UPDATE = ('update ovdc compute policies', '/cse/ovdc/%s/compute-policies')  # noqa: E501
    SYSTEM_INFO = ('get info of system', '/cse/system')
    SYSTEM_UPDATE = ('update system status', '/cse/system')
    TEMPLATE_LIST = ('list all templates', '/cse/templates')

    V35_OVDC_LIST = ('list ovdcs for v35', '/cse/3.0/ovdcs')
    V35_ORG_VDC_LIST = ('list org VDCs', '/cse/3.0/orgvdcs')
    V35_OVDC_INFO = ('get info of ovdc for v35', '/cse/3.0/ovdc/%s')
    V35_OVDC_UPDATE = ('enable or disable ovdc for a cluster kind for v35', '/cse/3.0/ovdc/%s', requests.codes.accepted)  # noqa: E501
    V35_TEMPLATE_LIST = ('list all v35 templates', '/cse/templates')

    PKS_CLUSTER_CONFIG = ('get config of PKS cluster', '/pks/clusters/%s/config')  # noqa: E501
    PKS_CLUSTER_CREATE = ('create PKS cluster', '/pks/clusters', requests.codes.accepted)  # noqa: E501
    PKS_CLUSTER_DELETE = ('delete PKS cluster', '/pks/cluster/%s', requests.codes.accepted)  # noqa: E501
    PKS_CLUSTER_INFO = ('get info of PKS cluster', '/pks/cluster/%s')
    PKS_CLUSTER_LIST = ('list PKS clusters', '/pks/clusters')
    PKS_CLUSTER_RESIZE = ('resize PKS cluster', '/pks/cluster/%s', requests.codes.accepted)  # noqa: E501
    PKS_OVDC_LIST = ('list all ovdcs', '/pks/ovdcs')
    PKS_ORG_VDC_LIST = ('list  org VDCs', '/pks/orgvdcs')
    PKS_OVDC_INFO = ('get info of the ovdc', '/pks/ovdc/%s')
    PKS_OVDC_UPDATE = ('enable or disable ovdc for pks', '/pks/ovdc/%s', requests.codes.accepted)  # noqa: E501


@unique
class ClusterMetadataKey(str, Enum):
    BACKWARD_COMPATIBLE_TEMPLATE_NAME = 'cse.template'
    CLUSTER_ID = 'cse.cluster.id'
    CSE_VERSION = 'cse.version'
    CONTROL_PLANE_IP = 'cse.master.ip'
    TEMPLATE_NAME = 'cse.template.name'
    TEMPLATE_REVISION = 'cse.template.revision'
    OS = "cse.os"
    DOCKER_VERSION = "cse.docker.version"
    KUBERNETES = "cse.kubernetes"
    KUBERNETES_VERSION = "cse.kubernetes.version"
    CNI = "cse.cni"
    CNI_VERSION = 'cse.cni.version'


@unique
class TemplateBuildKey(str, Enum):
    TEMPLATE_NAME = 'template_name'
    TEMPLATE_REVISION = 'template_revision'
    SOURCE_OVA_NAME = 'source_ova_name'
    SOURCE_OVA_HREF = 'source_ova_href'
    SOURCE_OVA_SHA256 = 'source_ova_sha256'
    ORG_NAME = 'org_name'
    VDC_NAME = 'vdc_name'
    CATALOG_NAME = 'catalog_name'
    CATALOG_ITEM_NAME = 'catalog_item_name'
    CATALOG_ITEM_DESCRIPTION = 'catalog_item_description'
    TEMP_VAPP_NAME = 'temp_vapp_name'
    TEMP_VM_NAME = 'temp_vm_name'
    CPU = 'cpu'
    MEMORY = 'memory'
    NETWORK_NAME = 'network_name'
    IP_ALLOCATION_MODE = 'ip_allocation_mode'
    STORAGE_PROFILE = 'storage_profile'
    CSE_PLACEMENT_POLICY = 'cse_placement_policy'
    REMOTE_COOKBOOK_VERSION = 'remote_template_cookbook_version'


@unique
class MQTTExtKey(str, Enum):
    API_FILTER_ID = 'api_filter_id'
    EXT_NAME = 'name'
    EXT_VERSION = 'version'
    EXT_VENDOR = 'vendor'
    EXT_PRIORITY = 'priority'
    EXT_ENABLED = 'enabled'
    EXT_AUTH_ENABLED = 'authorizationEnabled'
    EXT_DESCRIPTION = 'description'
    EXT_URN_ID = 'ext_urn_id'
    EXT_UUID = 'ext_uuid'
    EXT_LISTEN_TOPIC = 'listen_topic'
    EXT_RESPOND_TOPIC = 'respond_topic'


@unique
class MQTTExtTokenKey(str, Enum):
    TOKEN_NAME = 'name'
    TOKEN_TYPE = 'type'
    TOKEN_EXT_ID = 'extensionId'
    TOKEN = 'token'
    TOKEN_ID = 'token_id'


@unique
class ExtensionType(str, Enum):
    MQTT = 'MQTT'
    AMQP = 'AMQP'
    NONE = 'None'


@unique
class AclGrantType(str, Enum):
    MembershipACLGrant = "MembershipAccessControlGrant"


@unique
class AclAccessLevelId(str, Enum):
    AccessLevelReadOnly = "urn:vcloud:accessLevel:ReadOnly"
    AccessLevelReadWrite = "urn:vcloud:accessLevel:ReadWrite"
    AccessLevelFullControl = "urn:vcloud:accessLevel:FullControl"


@unique
class AclMemberId(str, Enum):
    SystemOrgId = "urn:vcloud:org:a93c9db9-7471-3192-8d09-a8f7eeda85f9"


@unique
class VappAccessKey(str, Enum):
    """Keys for VAPP access control."""

    IS_SHARED_TO_EVERYONE = 'isSharedToEveryone'
    ACCESS_SETTINGS = 'accessSettings'
    ACCESS_SETTING = 'accessSetting'


# CSE Service Role Name
CSE_SERVICE_ROLE_NAME = 'CSE Service Role'
CSE_SERVICE_ROLE_DESC = "CSE Service Role has all the rights necessary for \
        CSE to operate"
CSE_SERVICE_ROLE_RIGHTS = [
    "Access Control List: View",
    "Access Control List: Manage",
    "AMQP Settings: View",
    "Catalog: Add vApp from My Cloud",
    "Catalog: Create / Delete a Catalog",
    "Catalog: Edit Properties",
    "Catalog: Publish",
    "Catalog: Sharing",
    "Catalog: View ACL",
    "Catalog: View Private and Shared Catalogs",
    "Catalog: View Published Catalogs",
    "Content Library System Settings: View",
    "Custom entity: Create custom entity definitions",
    "Custom entity: Delete custom entity definitions",
    "Custom entity: Edit custom entity definitions",
    "Custom entity: View custom entity definitions",
    "Extension Services: View",
    "Extensions: View",
    "External Service: Manage",
    "External Service: View",
    "General: View Error Details",
    "Group / User: View",
    "Host: View",
    "Kerberos Settings: View",
    "Organization Network: View",
    "Organization vDC Compute Policy: Admin View",
    "Organization vDC Compute Policy: Manage",
    "Organization vDC Compute Policy: View",
    "Organization vDC Kubernetes Policy: Edit",
    "Organization vDC Network: Edit Properties",
    "Organization vDC Network: View Properties",
    "Organization vDC: Extended Edit",
    "Organization vDC: Extended View",
    "Organization vDC: View",
    "Organization: Perform Administrator Queries",
    "Organization: View",
    "Provider Network: View",
    "Provider vDC Compute Policy: Manage",
    "Provider vDC Compute Policy: View",
    "Provider vDC: View",
    "Right: Manage",
    "Right: View",
    "Rights Bundle: View",
    "Rights Bundle: Edit",
    "Role: Create, Edit, Delete, or Copy",
    "Service Configuration: Manage",
    "Service Configuration: View",
    "System Settings: View",
    "Task: Resume, Abort, or Fail",
    "Task: Update",
    "Task: View Tasks",
    "Token: Manage",
    "UI Plugins: Define, Upload, Modify, Delete, Associate or Disassociate",
    "UI Plugins: View",
    "vApp Template / Media: Copy",
    "vApp Template / Media: Create / Upload",
    "vApp Template / Media: Edit",
    "vApp Template / Media: View",
    "vApp Template: Checkout",
    "vApp Template: Import",
    "vApp: Allow All Extra Config",
    "vApp: Allow Ethernet Coalescing Extra Config",
    "vApp: Allow Latency Extra Config",
    "vApp: Allow Matching Extra Config",
    "vApp: Allow NUMA Node Affinity Extra Config",
    "vApp: Create / Reconfigure",
    "vApp: Delete",
    "vApp: Edit Properties",
    "vApp: Edit VM CPU and Memory reservation settings in all VDC types",
    "vApp: Edit VM CPU",
    "vApp: Edit VM Compute Policy",
    "vApp: Edit VM Hard Disk",
    "vApp: Edit VM Memory",
    "vApp: Edit VM Network",
    "vApp: Edit VM Properties",
    "vApp: Manage VM Password Settings",
    "vApp: Power Operations",
    "vApp: Shadow VM View",
    "vApp: Upload",
    "vApp: Use Console",
    "vApp: VM Boot Options",
    "vApp: VM Check Compliance",
    "vApp: VM Migrate, Force Undeploy, Relocate, Consolidate",
    "vApp: View VM and VM's Disks Encryption Status",
    "vApp: View VM metrics",
    "vCenter: View",
    "vSphere Server: View",
    "vmware:tkgcluster: Administrator Full access",
    "vmware:tkgcluster: Administrator View",
    "vmware:tkgcluster: Full Access",
    "vmware:tkgcluster: Modify",
    "vmware:tkgcluster: View"
]


@dataclass
class DefEntityPhase:
    """Supports two ways of creation.

    1. DefEntityPhase(DefEntityOperation.CREATE, DefEntityOperationStatus.SUCCEEDED) # noqa: E501
    2. DefEntityPhase.from_phase('CREATE:SUCCEEDED')
    """

    operation: DefEntityOperation
    status: DefEntityOperationStatus

    def __str__(self):
        return f'{self.operation}:{self.status}'

    @classmethod
    def from_phase(cls, phase: str):
        """Return instance of DefEntityPhase.

        :param str phase: defined entity phase value. ex: "CREATE:SUCCEEDED"
        :return: DefEntityPhase
        :rtype: <class DefEntityPhase>
        """
        operation, status = phase.split(':')
        return cls(DefEntityOperation[operation], DefEntityOperationStatus[status])  # noqa: E501

    def is_operation_status_success(self) -> bool:
        try:
            return self.status == DefEntityOperationStatus.SUCCEEDED
        except Exception:
            return False

    def is_entity_busy(self) -> bool:
        return self.status == DefEntityOperationStatus.IN_PROGRESS


@unique
class OvdcInfoKey(str, Enum):
    OVDC_NAME = 'name'
    ORG_NAME = 'org'
    K8S_PROVIDER = 'k8s provider'


@unique
class PKSOvdcInfoKey(str, Enum):
    PKS_API_SERVER = 'pks api server'
    AVAILABLE_PKS_PLANS = 'available pks plans'


# Network urn prefix
NETWORK_URN_PREFIX = 'urn:vcloud:network'

# Expose cluster name to be appended to cluster name for dnat rule
EXPOSE_CLUSTER_NAME_FRAGMENT = 'expose_dnat'


@unique
class VdcNetworkInfoKey(str, Enum):
    VALUES = 'values'
    CONNECTION = 'connection'
    ROUTER_REF = 'routerRef'
    NAME = 'name'


# Regex for ip: port
IP_PORT_REGEX = '[0-9]+(?:\\.[0-9]+){3}(:[0-9]+)?'

# Default first page
DEFAULT_FIRST_PAGE = 1

# Nat rules pagination constants
NAT_DEFAULT_PAGE_SIZE = 25

# Pagination constants for used IP addresses
USED_IP_ADDRESS_PAGE_SIZE = 10

# Context headers
TENANT_CONTEXT_HEADER = 'X-VMWARE-VCLOUD-TENANT-CONTEXT'
AUTH_CONTEXT_HEADER = 'X-VMWARE-VCLOUD-AUTH-CONTEXT'
VCLOUD_AUTHORIZATION_HEADER = 'X-vCloud-Authorization'


@unique
class PostCustomizationStatus(Enum):
    NONE = None
    IN_PROGRESS = 'in_progress'
    SUCCESSFUL = 'successful'


POST_CUSTOMIZATION_SCRIPT_EXECUTION_STATUS = 'post_customization_script_execution_status'  # noqa: E501
POST_CUSTOMIZATION_SCRIPT_EXECUTION_FAILURE_REASON = 'post_customization_script_execution_failure_reason'  # noqa: E501
DEFAULT_POST_CUSTOMIZATION_STATUS_LIST = [cust_status.value for cust_status in PostCustomizationStatus]  # noqa: E501
DEFAULT_POST_CUSTOMIZATION_POLL_SEC = 5
DEFAULT_POST_CUSTOMIZATION_TIMEOUT_SEC = 180



# coding: utf-8

"""
    TKG Kubernetes API

    This API provides to vCD tenants the means to provision (create and update) Tanzu Kubernetes Grid clusters. This is complementary to the defined-entity APIs:    GET /cloudapi/1.0.0/entities/urn:vcloud:entity:vmware.tkgcluster:1.0.0:{id} which allows to retrieve the clusters created by the API presented here. This is why you will not find here a GET operation for the corresponding entity.   # noqa: E501

    OpenAPI spec version: 1.0.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from container_service_extension.client.tkgclient.models.tkg_cluster_spec_distribution import TkgClusterSpecDistribution  # noqa: F401,E501
from container_service_extension.client.tkgclient.models.tkg_cluster_spec_settings import TkgClusterSpecSettings  # noqa: F401,E501
from container_service_extension.client.tkgclient.models.tkg_cluster_spec_topology import TkgClusterSpecTopology  # noqa: F401,E501


class TkgClusterSpec(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'distribution': 'TkgClusterSpecDistribution',
        'settings': 'TkgClusterSpecSettings',
        'topology': 'TkgClusterSpecTopology'
    }

    attribute_map = {
        'distribution': 'distribution',
        'settings': 'settings',
        'topology': 'topology'
    }

    def __init__(self, distribution=None, settings=None, topology=None):  # noqa: E501
        """TkgClusterSpec - a model defined in Swagger"""  # noqa: E501

        self._distribution = None
        self._settings = None
        self._topology = None
        self.discriminator = None

        self.distribution = distribution
        if settings is not None:
            self.settings = settings
        self.topology = topology

    @property
    def distribution(self):
        """Gets the distribution of this TkgClusterSpec.  # noqa: E501


        :return: The distribution of this TkgClusterSpec.  # noqa: E501
        :rtype: TkgClusterSpecDistribution
        """
        return self._distribution

    @distribution.setter
    def distribution(self, distribution):
        """Sets the distribution of this TkgClusterSpec.


        :param distribution: The distribution of this TkgClusterSpec.  # noqa: E501
        :type: TkgClusterSpecDistribution
        """
        if distribution is None:
            raise ValueError("Invalid value for `distribution`, must not be `None`")  # noqa: E501

        self._distribution = distribution

    @property
    def settings(self):
        """Gets the settings of this TkgClusterSpec.  # noqa: E501


        :return: The settings of this TkgClusterSpec.  # noqa: E501
        :rtype: TkgClusterSpecSettings
        """
        return self._settings

    @settings.setter
    def settings(self, settings):
        """Sets the settings of this TkgClusterSpec.


        :param settings: The settings of this TkgClusterSpec.  # noqa: E501
        :type: TkgClusterSpecSettings
        """

        self._settings = settings

    @property
    def topology(self):
        """Gets the topology of this TkgClusterSpec.  # noqa: E501


        :return: The topology of this TkgClusterSpec.  # noqa: E501
        :rtype: TkgClusterSpecTopology
        """
        return self._topology

    @topology.setter
    def topology(self, topology):
        """Sets the topology of this TkgClusterSpec.


        :param topology: The topology of this TkgClusterSpec.  # noqa: E501
        :type: TkgClusterSpecTopology
        """
        if topology is None:
            raise ValueError("Invalid value for `topology`, must not be `None`")  # noqa: E501

        self._topology = topology

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, TkgClusterSpec):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# container-service-extension
# Copyright (c) 2020 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause
from typing import Dict, List, Optional

import lxml
import pyvcloud.vcd.client as vcd_client
import pyvcloud.vcd.utils as pyvcloud_utils
import pyvcloud.vcd.vapp as vcd_vapp

import container_service_extension.common.constants.server_constants as server_constants  # noqa: E501
import container_service_extension.common.constants.shared_constants as shared_constants  # noqa: E501
import container_service_extension.common.utils.core_utils as utils
import container_service_extension.common.utils.pyvcloud_utils as vcd_utils
import container_service_extension.lib.cloudapi.constants as cloudapi_constants
import container_service_extension.rde.common.entity_service as def_entity_svc
import container_service_extension.rde.constants as def_constants
import container_service_extension.rde.models.common_models as common_models


class ClusterACLService:
    """Manages retrieving and setting Cluster ACL information."""

    def __init__(self, cluster_id: str,
                 client: vcd_client.Client):
        self._client = client
        self._cloudapi_client = \
            vcd_utils.get_cloudapi_client_from_vcd_client(client)
        self._cluster_id = cluster_id
        self._def_entity: Optional[common_models.DefEntity] = None
        self._vapp: Optional[vcd_vapp.VApp] = None

    @property
    def def_entity(self):
        if self._def_entity is None:
            entity_svc = def_entity_svc.DefEntityService(self._cloudapi_client)
            self._def_entity = entity_svc.get_tkg_or_def_entity(self._cluster_id)  # noqa: E501
        return self._def_entity

    @property
    def vapp(self):
        if self._vapp is None:
            self._vapp = vcd_vapp.VApp(self._client,
                                       href=self.def_entity.externalId)
        return self._vapp

    def get_def_entity_acl_response(self, page, page_size):
        acl_path = f'{cloudapi_constants.CloudApiResource.ENTITIES}/' \
                   f'{self._cluster_id}/' \
                   f'{cloudapi_constants.CloudApiResource.ACL}' \
                   f'?{shared_constants.PaginationKey.PAGE_NUMBER}={page}&' \
                   f'{shared_constants.PaginationKey.PAGE_SIZE}={page_size}'
        de_acl_response = self._cloudapi_client.do_request(
            method=shared_constants.RequestMethod.GET,
            cloudapi_version=cloudapi_constants.CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=acl_path)
        return de_acl_response

    def list_def_entity_acl_entries(self):
        """List def entity acl.

        :return: Generator of cluster acl entries
        :rtype: Generator[ClusterAclEntry]
        """
        page_num = 0
        while True:
            page_num += 1
            response_body = self.get_def_entity_acl_response(
                page_num, shared_constants.CSE_PAGINATION_DEFAULT_PAGE_SIZE)
            values = response_body[shared_constants.PaginationKey.VALUES]
            if len(values) == 0:
                break
            for acl_entry in values:
                yield common_models.ClusterAclEntry(**acl_entry)

    def create_user_id_to_acl_entry_dict(self):
        """Get all def entity acl values from all pages.

        :return dict of user id keys and def_models.ClusterAclEntry values
        """
        user_id_to_acl_entry = {}
        for acl_entry in self.list_def_entity_acl_entries():
            user_id_to_acl_entry[acl_entry.memberId] = acl_entry
        return user_id_to_acl_entry

    def share_def_entity(self, acl_entry: common_models.ClusterAclEntry):
        access_controls_path = \
            f'{cloudapi_constants.CloudApiResource.ENTITIES}/' \
            f'{self._cluster_id}/{cloudapi_constants.CloudApiResource.ACL}'
        ent_kind = self.def_entity.entity.kind \
            if hasattr(self.def_entity, 'entity') else self.def_entity.kind
        if ent_kind in \
                [shared_constants.ClusterEntityKind.NATIVE.value,
                 shared_constants.ClusterEntityKind.TKG_PLUS.value]:
            org_id = vcd_utils.extract_id(self.def_entity.org.id)
        elif ent_kind == shared_constants.ClusterEntityKind.TKG_S.value:
            vdc_name = self.def_entity.metadata.virtualDataCenterName
            org_id = vcd_utils.get_org_id_from_vdc_name(
                client=self._client,
                vdc_name=vdc_name)
        else:
            raise Exception(f"Invalid entity kind: {ent_kind}")

        payload = acl_entry.construct_filtered_dict(
            include=shared_constants.DEF_ENTITY_ACCESS_CONTROL_KEYS)
        self._cloudapi_client.do_request(
            method=shared_constants.RequestMethod.POST,
            cloudapi_version=cloudapi_constants.CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=access_controls_path,
            additional_request_headers={server_constants.TENANT_CONTEXT_HEADER: org_id},  # noqa: E501
            payload=payload)

    def unshare_def_entity(self, acl_id):
        delete_path = f'{cloudapi_constants.CloudApiResource.ENTITIES}/' \
            f'{self._cluster_id}/{cloudapi_constants.CloudApiResource.ACL}/' \
            f'{acl_id}'
        self._cloudapi_client.do_request(
            method=shared_constants.RequestMethod.DELETE,
            cloudapi_version=cloudapi_constants.CloudApiVersion.VERSION_1_0_0,
            resource_url_relative_path=delete_path)

    def update_native_def_entity_acl(self, update_acl_entries: List[common_models.ClusterAclEntry],  # noqa: E501
                                     prev_user_id_to_acl_entry: Dict[str, common_models.ClusterAclEntry]):  # noqa: E501
        """Update native defined entity acl.

        :param list update_acl_entries: list of def_models.ClusterAclEntry
        :param dict prev_user_id_to_acl_entry: dict mapping user id to
            def_models.ClusterAclEntry

        :return: dictionary of memberId keys and access level values
        """
        own_prev_user_id_to_acl_entry = prev_user_id_to_acl_entry.copy()

        # Share defined entity
        ent_org_user_id_names_dict = vcd_utils.create_org_user_id_to_name_dict(
            client=self._client,
            org_name=self.def_entity.org.name)
        user_acl_level_dict = {}
        share_acl_entry = common_models.ClusterAclEntry(
            grantType=shared_constants.MEMBERSHIP_GRANT_TYPE,
            memberId=None,
            accessLevelId=None)
        for acl_entry in update_acl_entries:
            user_id = acl_entry.memberId
            if ent_org_user_id_names_dict.get(user_id) is None:
                # This user must be from the system org and does not need
                # to be re-shared. Sharing can not happen from a tenant user
                # to a system user.
                del own_prev_user_id_to_acl_entry[user_id]
                continue
            acl_level = acl_entry.accessLevelId
            share_acl_entry.memberId = user_id
            share_acl_entry.accessLevelId = acl_level
            user_acl_level_dict[user_id] = acl_level
            self.share_def_entity(share_acl_entry)

            # Remove entry from previous user acl info
            if own_prev_user_id_to_acl_entry.get(user_id):
                del own_prev_user_id_to_acl_entry[user_id]

        # Delete def entity acl entries not in update_acl_entries
        for _, acl_entry in own_prev_user_id_to_acl_entry.items():
            self.unshare_def_entity(acl_entry.id)
        return user_acl_level_dict

    def native_get_vapp_settings_only_vapp_shared(self, def_entity_user_ids: set):  # noqa: E501
        """Get vapp settings in which the defined entity is not also shared.

        :param set def_entity_user_ids: set of user ids that have access to
            the defined entity

        :return: vapp settings in which only the vapp is shared
        :rtype: list
        """
        non_updated_access_settings = []
        vapp_access_settings: lxml.objectify.ObjectifiedElement = \
            self.vapp.get_access_settings()
        # Only add entries in which the defined entity is not shared
        if hasattr(vapp_access_settings, 'AccessSettings'):
            vapp_access_settings_attr = vapp_access_settings.AccessSettings
            for child_obj in vapp_access_settings_attr.getchildren():
                # Get user_urn
                child_obj_attrib = child_obj.getchildren()[0].attrib
                shared_href = child_obj_attrib.get('href')
                user_id = utils.extract_id_from_href(shared_href)
                user_urn = f'{shared_constants.USER_URN_PREFIX}{user_id}'

                # Add entries in which only vapp is shared
                if user_urn not in def_entity_user_ids:
                    user_name = child_obj_attrib.get('name')

                    curr_setting = form_vapp_access_setting_entry(
                        access_level=str(child_obj.AccessLevel),
                        name=user_name,
                        href=shared_href,
                        user_id=user_id)
                    non_updated_access_settings.append(curr_setting)
        return non_updated_access_settings

    def native_update_vapp_access_settings(self, prev_user_id_to_acl_entry_dict,  # noqa: E501
                                           update_cluster_acl_entries: List[
                                               common_models.ClusterAclEntry]):
        def_entity_user_ids = {acl_entry.memberId for _, acl_entry in
                               prev_user_id_to_acl_entry_dict.items()}
        total_vapp_access_settings = self.native_get_vapp_settings_only_vapp_shared(def_entity_user_ids)  # noqa: E501

        # Add updated access settings
        vapp_access_settings: lxml.objectify.ObjectifiedElement = \
            self.vapp.get_access_settings()
        api_uri = self._client.get_api_uri()
        system_user_names: Optional[set] = None
        if self._client.is_sysadmin():
            system_user_names = vcd_utils.get_org_user_names(
                client=self._client,
                org_name=shared_constants.SYSTEM_ORG_NAME)
        for acl_entry in update_cluster_acl_entries:
            user_name = acl_entry.username
            # Skip system users since sharing can't be outside an org
            if system_user_names and user_name in system_user_names:
                continue
            user_id = pyvcloud_utils.extract_id(acl_entry.memberId)
            access_level = pyvcloud_utils.extract_id(acl_entry.accessLevelId)

            # Use 'Change' instead of 'ReadWrite' for vApp access level
            if access_level == shared_constants.READ_WRITE:
                access_level = server_constants.CHANGE_ACCESS
            user_setting = form_vapp_access_setting_entry(
                access_level=access_level,
                name=user_name,
                href=f'{api_uri}{server_constants.ADMIN_USER_PATH}{user_id}',
                user_id=user_id)
            total_vapp_access_settings.append(user_setting)

        vapp_share_contents = {
            server_constants.VappAccessKey.IS_SHARED_TO_EVERYONE:
                bool(vapp_access_settings.IsSharedToEveryone),
            server_constants.VappAccessKey.ACCESS_SETTINGS:
                {server_constants.VappAccessKey.ACCESS_SETTING: total_vapp_access_settings}  # noqa: E501
        }

        org_id = pyvcloud_utils.extract_id(self.def_entity.org.id)
        org_name = self.def_entity.org.name
        extra_vapp_headers = {
            server_constants.TENANT_CONTEXT_HEADER: org_id,
            server_constants.AUTH_CONTEXT_HEADER: org_name,
            server_constants.VCLOUD_AUTHORIZATION_HEADER: org_name
        }
        self._client.post_resource(
            uri=f'{self.vapp.href}{def_constants.ACTION_CONTROL_ACCESS_PATH}',
            contents=vapp_share_contents,
            media_type='application/*+json',
            extra_headers=extra_vapp_headers)

    def get_cluster_entity(self):
        return self.def_entity


def form_vapp_access_setting_entry(access_level, name, href, user_id):
    vapp_access_setting = {
        shared_constants.AccessControlKey.ACCESS_LEVEL: access_level,
        shared_constants.AccessControlKey.SUBJECT: {
            shared_constants.AccessControlKey.NAME: name,
            shared_constants.AccessControlKey.HREF: href,
            shared_constants.AccessControlKey.ID: user_id
        }
    }
    return vapp_access_setting



# container-service-extension
# Copyright (c) 2017 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import pyvcloud.vcd.client as vcd_client

from container_service_extension.client.de_cluster import DECluster
from container_service_extension.client.de_cluster_native import DEClusterNative  # noqa: E501
from container_service_extension.client.de_cluster_tkg_s import DEClusterTKGS
from container_service_extension.client.legacy_cluster_native import LegacyClusterNative  # noqa: E501
from container_service_extension.common.constants.shared_constants import ClusterEntityKind  # noqa: E501


class Cluster:
    """Returns the cluster class as determined by API version."""

    def __new__(cls, client: vcd_client.Client, k8_runtime=None):
        """Create the right cluster class for the negotiated API version.

        In case of ApiVersion.VERSION_35, return specific instance if the
        cluster kind is known up-front.

        If the cluster entity kind is unknown, return instance of
        DefEntityCluster for all common operations like get_cluster_info(),
        list_clusters()

        :param pyvcloud.vcd.client client: vcd client

        :return: instance of version specific client side cluster
        """
        api_version = client.get_api_version()
        if float(api_version) < float(vcd_client.ApiVersion.VERSION_35.value):   # noqa: E501
            return LegacyClusterNative(client)
        elif float(api_version) >= float(vcd_client.ApiVersion.VERSION_35.value):  # noqa: E501
            if k8_runtime == ClusterEntityKind.NATIVE.value or k8_runtime == ClusterEntityKind.TKG_PLUS.value:  # noqa: E501
                return DEClusterNative(client)
            elif k8_runtime == ClusterEntityKind.TKG_S.value:
                return DEClusterTKGS(client)
            else:
                return DECluster(client)



# container-service-extension
# Copyright (c) 2021 VMware, Inc. All Rights Reserved.
# SPDX-License-Identifier: BSD-2-Clause

import functools

from container_service_extension.common.constants.shared_constants import RequestKey  # noqa: E501
import container_service_extension.exception.exceptions as cse_exception
from container_service_extension.lib.cloudapi.cloudapi_client import \
    CloudApiClient
from container_service_extension.logging.logger import SERVER_LOGGER as LOGGER
from container_service_extension.rde.backend import cluster_service_factory
import container_service_extension.rde.constants as rde_constants
import container_service_extension.rde.utils as rde_utils
import container_service_extension.rde.validators.validator_factory as rde_validator_factory  # noqa: E501
from container_service_extension.security.context.behavior_request_context import RequestContext  # noqa: E501


# Responsibility of the functions in this file:
# 1. Validate the Input payload based on the (BehaviorOperation, payload_version).  # noqa: E501
# Get the validator based on the payload_version.
# 2. Based on the rde_in_use, convert the entity if necessary
# 3. Based on the rde_in_use, call the appropriate backend method.
# 4. Return the success_payload_string.

def exception_handler(func):
    """Decorate to trap exceptions and process them.

    Raise errors of type KeyError, TypeError, ValueError as
    BadRequestError.

    Also raises BadRequest and Internal Server Errors from backend.

    :param method func: decorated function

    :return: reference to the function that executes the decorated function
        and traps exceptions raised by it.
    """
    @functools.wraps(func)
    def exception_handler_wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except (KeyError, TypeError, ValueError) as error:
            LOGGER.error(error, exc_info=True)
            raise cse_exception.BadRequestError(error_message=str(error))
        except Exception as error:
            LOGGER.error(error, exc_info=True)
            if not isinstance(error, cse_exception.CseRequestError):
                raise cse_exception.InternalServerRequestError(error_message=str(error))  # noqa: E501
            raise error
        return result
    return exception_handler_wrapper


@exception_handler
def create_cluster(behavior_ctx: RequestContext):
    entity_id: str = behavior_ctx.entity_id
    input_entity: dict = behavior_ctx.entity
    cloudapi_client: CloudApiClient = behavior_ctx.op_ctx.cloudapi_client

    payload_version: str = input_entity.get(rde_constants.PayloadKey.PAYLOAD_VERSION)  # noqa: E501
    rde_utils.raise_error_if_unsupported_payload_version(payload_version)

    # Validate the Input payload based on the (Operation, payload_version).
    # Get the validator based on the payload_version
    input_rde_version = rde_constants.MAP_INPUT_PAYLOAD_VERSION_TO_RDE_VERSION[payload_version]  # noqa: E501
    rde_validator_factory.get_validator(
        rde_version=input_rde_version). \
        validate(cloudapi_client=cloudapi_client, entity=input_entity)

    # Convert the input entity to runtime rde format.
    # Based on the runtime rde, call the appropriate backend method.
    converted_input_entity = rde_utils.convert_input_rde_to_runtime_rde_format(input_entity)  # noqa: E501
    svc = cluster_service_factory.ClusterServiceFactory(behavior_ctx).get_cluster_service()  # noqa: E501
    return svc.create_cluster(entity_id=entity_id, input_native_entity=converted_input_entity)  # noqa: E501


@exception_handler
def update_cluster(behavior_ctx: RequestContext):
    entity_id: str = behavior_ctx.entity_id
    input_entity: dict = behavior_ctx.entity
    cloudapi_client: CloudApiClient = behavior_ctx.op_ctx.cloudapi_client

    payload_version: str = input_entity.get(rde_constants.PayloadKey.PAYLOAD_VERSION)  # noqa: E501
    rde_utils.raise_error_if_unsupported_payload_version(
        payload_version=payload_version)

    # Validate the Input payload based on the (Operation, payload_version).
    # Get the validator based on the payload_version
    input_rde_version = rde_constants.MAP_INPUT_PAYLOAD_VERSION_TO_RDE_VERSION[payload_version]  # noqa: E501
    rde_validator_factory.get_validator(
        rde_version=input_rde_version). \
        validate(cloudapi_client=cloudapi_client, entity=input_entity)

    # Convert the input entity to runtime rde format.
    # Based on the runtime rde, call the appropriate backend method.
    converted_input_entity = rde_utils.convert_input_rde_to_runtime_rde_format(input_entity)  # noqa: E501
    svc = cluster_service_factory.ClusterServiceFactory(behavior_ctx).get_cluster_service()  # noqa: E501
    return svc.update_cluster(entity_id, input_native_entity=converted_input_entity)  # noqa: E501


@exception_handler
def delete_cluster(behavior_ctx: RequestContext):
    entity_id: str = behavior_ctx.entity_id

    svc = cluster_service_factory.ClusterServiceFactory(behavior_ctx).get_cluster_service()  # noqa: E501

    return svc.delete_cluster(entity_id)


@exception_handler
def get_kubeconfig(behavior_ctx: RequestContext):
    cluster_id: str = behavior_ctx.entity_id
    svc = cluster_service_factory.ClusterServiceFactory(behavior_ctx).get_cluster_service()  # noqa: E501
    return svc.get_cluster_config(cluster_id)


@exception_handler
def nfs_node_delete(behavior_ctx: RequestContext):
    entity_id: str = behavior_ctx.entity_id
    node_to_delete: str = behavior_ctx.arguments.get(RequestKey.NODE_NAME)
    svc = cluster_service_factory.ClusterServiceFactory(behavior_ctx).get_cluster_service()  # noqa: E501
    return svc.delete_nodes(entity_id, nodes_to_del=[node_to_delete])
